#include "board.hpp"
#include "state.hpp"
#include "communication.hpp"
#include "position.hpp"
#include "ground_truth.hpp"
#include <iomanip>
#include <set>
#include <algorithm>
using namespace std;

robot_position robot;
ground_truth gt;
cylinder_config cylinders;
ball_positions balls;


/*
int pos_x(double angle, double orientation, int distance, int x_offset) {
    return sin(angle+orientation)*distance+x_offset;
}

int pos_y(double angle, double orientation, int distance, int y_offset) {
    return cos(angle+orientation)*distance+y_offset;
}
*/

void hard_code_add(corn_draw& input, int y1, int x1, int y2, int x2, short index)
{
    CvPoint temp;
    temp.y=y1;
    temp.x=x1;
    input.lower=temp;
    temp.y=y2;
    temp.x=x2;
    input.upper=temp;
    input.color='u';
    input.index=index;
}

typedef struct
{
    double value; // will be <29 bit @ 1024x768, if all pixels are 100% white. (not likely)
    int count;
} pixel_accumulator;

bool descending_cmp(const double &a, const double &b)
{
    return a>b;
}


void init_game_elements()
{
    vector<CvPoint> corn_positions;
    corn_populate(corn_positions);
    for (int i=0; i<corn_positions.size(); i++)
    {
        cylinders.cylinder[i].y=corn_positions[i].y;
        cylinders.cylinder[i].x=corn_positions[i].x;
    }
    vector<CvPoint> tomato_positions;
    tomato_populate(tomato_positions);
    for (int i=0; i<tomato_positions.size(); i++)
    {
        balls.tomato[i].y=tomato_positions[i].y;
        balls.tomato[i].x=tomato_positions[i].x;
    }
    for (int i=0; i<12; i++)
    {
        balls.orange[i].y=-1;
        balls.orange[i].x=-1;
    }
}

// finding if something on the ground is within view area or not.
// fastest is to draw a triangle, and use a scanline to figure out how many lines it intersects before hitting the point.

// since speed *REALLY* isn't an issue, we are just calculating if it is in the image or not.
// this assumes that the ground truth can be calibrated properly.
bool in_view_area(robot_position &pos, int y, int x)
{
    robot_position left_camera;
    robot_position right_camera;
    camera_positions(pos,left_camera,right_camera);
    int rel_y = y-left_camera.y;
    int rel_x = x-left_camera.x;
    CvPoint left_view = gt.image_position_from_real_coord(rel_y,rel_x, pos.psi, 0);
    rel_y = y-right_camera.y;
    rel_x = x-right_camera.x;
    CvPoint right_view = gt.image_position_from_real_coord(rel_y,rel_x, pos.psi, 0);
    if (left_view.y < 768 && right_view.y < 768)
    {
        if (left_view.x > 0 && left_view.x < 1024 && right_view.x > 0 && right_view.x < 1024)
        {
            return true;
        }
        else
        {
            return false;
        }
    }
    else
    {
        return false;
    }
    return false;
}

CvPoint rotate_axis_single(robot_position &pos, double y, double x)
{
    return position_from_distance_and_orientation(pos.psi,y,x);
}


CvPoint rotate_axis(robot_position &pos, CvPoint2D32f input)
{
    robot_position left_camera;
    robot_position right_camera;
    robot_position middle;
    camera_positions(pos,left_camera,right_camera);
    middle.y=(left_camera.y+right_camera.y)/2; // these are equal, so not really necessary to take the middle.
    middle.x=(left_camera.x+right_camera.x)/2;
    middle.psi=pos.psi;
    return rotate_axis_single(middle, input.y*10.0,input.x*10.0);
}

double object_distance(coord &a, CvPoint &b)
{
    int rel_x = a.x-b.x;
    int rel_y = a.y-b.y;
    return sqrt(rel_x*rel_x+rel_y*rel_y);
}

void update_general(vector< CvPoint2D32f > &objects,set<int> &indexes,coord *object_list, int size)
{
    // for each object - check if present.
    set<int>::iterator it;
    for (int i=0; i<objects.size(); i++)
    {
        CvPoint p = rotate_axis(robot,objects.at(i));
        cout << "OBJECT_UPDATE_GENERAL Y: " << p.y << " X: " << p.x << endl;
        if (p.y < 0 || p.y > 2100 || p.x < 0 || p.x > 3000)
        {
            continue;
        }
        double min_distance=99999999.0;
        int object_id=-1;
        for (it=indexes.begin(); it != indexes.end(); ++it)
        {
            if (object_list[*it].y == -1 && object_list[*it].x == -1)
            {
                continue;
            }
            if (object_distance(object_list[*it],p) < min_distance)
            {
                min_distance = object_distance(object_list[*it],p);
                object_id = *it;
            }
        }
        if (object_id != -1)
        {
            printf("OBJECT ID: %d, Y: %d, X: %d, (Y: %d, X: %d)\n",*it,p.y,p.x,object_list[object_id].y,object_list[object_id].x);
            printf("OBJECTS LEFT: %d\n", (int) indexes.size());
            indexes.erase(object_id);
            object_list[object_id].y=p.y;
            object_list[object_id].x=p.x;
        }
        else
        {
            int count=0;
            it=indexes.begin();
            if (it != indexes.end())
            {
                printf("OBJECT ID: %d, Y: %d, X: %d, (Y: %d, X: %d)\n",*it,p.y,p.x,object_list[*it].y,object_list[*it].x);
                object_list[*it].y = p.y;
                object_list[*it].x = p.x;
                indexes.erase(it);
            }
            else
            {
                // out of non-present balls - starting to take best matching non-visible balls

            }
            //printf("BALL_IN_VIEW less than BALLS_FOUND - index placement postponed\n");
        }
    }

}

// this routine is used to update the game element map for sending to the AI
void update_elements(uint64_t timestamp, vector< CvPoint2D32f > &tomatoes, vector< CvPoint2D32f > &oranges, vector< CvPoint2D32f > &corns)
{
    robot = getOurPosInterpolated(timestamp);
    //robot.psi=90*M_PI/180.0;
    robot_position left_camera;
    robot_position right_camera;
    camera_positions(robot,left_camera,right_camera);
    //printf("Y: Axis: %d, Left: %d, Right: %d\n",robot.y,left_camera.y,right_camera.y);
    //printf("X: Axis: %d, Left: %d, Right: %d\n",robot.x,left_camera.x,right_camera.x);
    //robot.psi = 0*M_PI/180;
    set<int> cylinder_indexes;
    set<int> tomato_indexes;
    set<int> orange_indexes;
    set<int>::iterator it;
    // checking only white cylinders.
    for (int i=0; i<18; i++)
    {
        if (cylinders.cylinder[i].color == 'w')
        {
            cylinder_indexes.insert(i);
        }
    }
    // checking all tomatoes
    for (int i=0; i<14; i++)
    {
        tomato_indexes.insert(i);
    }
    // checking all oranges
    for (int i=0; i<12; i++)
    {
        orange_indexes.insert(i);
    }

    // checking if element is within radius of robot.
    // also checking information based on lens geometry.
    const int radius = 350; // given in mm
    for (it=cylinder_indexes.begin(); it != cylinder_indexes.end(); ++it)
    {
        int y = cylinders.cylinder[*it].y;
        int x = cylinders.cylinder[*it].x;
        if (!in_view_area(robot,y,x) || (y == -1 && x == -1))
        {
            cylinder_indexes.erase(it);
        }
    }
    for (it=tomato_indexes.begin(); it != tomato_indexes.end(); ++it)
    {
        int y = balls.tomato[*it].y;
        int x = balls.tomato[*it].x;
        if (in_view_area(robot,y,x))
        {
            printf("IN_VIEW_AREA: %d %d\n",y,x);
        }
        if (!in_view_area(robot,y,x) && (y != -1 && x != -1))
        {
            tomato_indexes.erase(it);
        }
    }

    for (it=orange_indexes.begin(); it != orange_indexes.end(); ++it)
    {
        int y = balls.orange[*it].y;
        int x = balls.orange[*it].x;
        if (!in_view_area(robot,y,x) && (y != -1 && x != -1))
        {
            orange_indexes.erase(it);
        }
    }
    cout << "UPDATE_ELEMENTS: tomatoes: " << tomato_indexes.size() << " oranges: " << orange_indexes.size() << " cylinders: " << cylinder_indexes.size() << endl;
    // remove all elements that are in view area (for testing)

    // do update_general here
    update_general(tomatoes,tomato_indexes, balls.tomato, 14);
    update_general(oranges, orange_indexes, balls.orange,12);
    cout << "ORANGES: " << oranges.size() << endl;
    //update_general(corns, cylinder_indexes, cylinders.cylinder,18);
    printf("--------------------------------------------------------------\n");
    for (int i=0; i<14; i++)
    {
        printf("SEND_TOMATO_POS: %d: Y: %d X: %d\n",i,balls.tomato[i].y,balls.tomato[i].x);
    }
    for (int i=0; i<12; i++)
    {
        printf("SEND_ORANGE_POS: %d: Y: %d X: %d\n",i,balls.orange[i].y,balls.orange[i].x);
    }





//   if(tomato_indexes.size() > 0 || orange_indexes.size() > 0) {
    for (it=tomato_indexes.begin(); it != tomato_indexes.end(); ++it)
    {

        printf("BALL_IN_VIEW (%d,%d,%f): Y: %d, X: %d\n",robot.y,robot.x,robot.psi,balls.tomato[*it].y,balls.tomato[*it].x);
        balls.tomato[*it].y=-1;
        balls.tomato[*it].x=-1;
    }
    for (it=orange_indexes.begin(); it != orange_indexes.end(); ++it)
    {
        balls.orange[*it].y=-1;
        balls.orange[*it].x=-1;
    }
    balls.counter=getPickupCount();
    send_ball_positions(balls);
//    }
    if (cylinder_indexes.size() > 0)
    {
        for (it=cylinder_indexes.begin(); it != cylinder_indexes.end(); ++it)
        {
            cylinders.cylinder[*it].y=-1;
            cylinders.cylinder[*it].x=-1;
        }
        cylinders.counter=getPickupCount();
        //send_cylinder_positions(cylinders);
    }

}

void draw_cylinder_config(RgbImage& input) {
    vector<corn_draw> corn;
    corn_draw temp;
    corn.assign(18,temp);
    if (our_color == 0)
    {
        hard_code_add(corn[0],651,904,441,970,1);
        hard_code_add(corn[1],384,795,268,820,2);
        hard_code_add(corn[2],292,757,210,771,3);
        hard_code_add(corn[3],436,403,298,391,4);
        hard_code_add(corn[4],315,504,226,504,5);
        hard_code_add(corn[5],259,552,192,552,6);
        hard_code_add(corn[6],187,340,241,166,7);
        hard_code_add(corn[7],274,328,202,321,8);
        hard_code_add(corn[8],289,69,213,48,9);
        hard_code_add(corn[9],243,213,184,202,10);
        for (int i=0; i<10; i++)
        {
            cvLine(input.raw(), corn[i].lower,corn[i].upper,cvScalar(10*(i+10),0,0),15,8,0);
        }
    }
    else if (our_color == 1)
    {
        hard_code_add(corn[8],280,913,208,930,9);
        hard_code_add(corn[9],243,775,184,780,10);
        hard_code_add(corn[10],334,816,235,831,11);
        hard_code_add(corn[11],273,664,204,672,12);
        hard_code_add(corn[12],436,613,297,624,13);
        hard_code_add(corn[13],319,493,229,493,14);
        hard_code_add(corn[14],265,441,199,438,15);
        hard_code_add(corn[15],690,124,468,46,16);
        hard_code_add(corn[16],393,196,277,175,17);
        hard_code_add(corn[17],297,226,220,216,18);
        for (int i=8; i<corn.size(); i++)
        {
            cvLine(input.raw(), corn[i].lower,corn[i].upper,cvScalar(10*i,0,0),15,8,0);
        }
    }
}

// this routine is used to find the cylinder configuration from the start position on the playing field.
// It assumes the robot is at a specific x,y coordinate, with a specific orientation.
// (This is achieved by using a shape in each corner, that the robot is placed against,
// so that it gets the same position every time.
void hard_coded_cylinders(RgbImage& input,BwImage& regions)
{
    vector<corn_draw> corn;
    corn_draw temp;
    corn.assign(18,temp);
    if (our_color == 0)
    {
        // BLUE
        hard_code_add(corn[0],651,904,441,970,1);
        hard_code_add(corn[1],384,795,268,820,2);
        hard_code_add(corn[2],292,757,210,771,3);
        hard_code_add(corn[3],436,403,298,391,4);
        hard_code_add(corn[4],315,504,226,504,5);
        hard_code_add(corn[5],259,552,192,552,6);
        hard_code_add(corn[6],187,340,241,166,7);
        hard_code_add(corn[7],274,328,202,321,8);
        hard_code_add(corn[8],289,69,213,48,9);
        hard_code_add(corn[9],243,213,184,202,10);
        for (int i=0; i<10; i++)
        {
            cvLine(regions.raw(), corn[i].lower,corn[i].upper,cvScalar(corn[i].index),15,8,0);
        }
    }
    else if (our_color == 1)
    {
        // YELLOW
	// check image
        hard_code_add(corn[8],280,913,208,930,9);
        hard_code_add(corn[9],243,775,184,780,10);
        hard_code_add(corn[10],334,816,235,831,11);
        hard_code_add(corn[11],273,664,204,672,12);
        hard_code_add(corn[12],436,613,297,624,13);
        hard_code_add(corn[13],319,493,229,493,14);
        hard_code_add(corn[14],265,441,199,438,15);
        hard_code_add(corn[15],690,124,468,46,16);
        hard_code_add(corn[16],393,196,277,175,17);
        hard_code_add(corn[17],297,226,220,216,18);
        for (int i=8; i<corn.size(); i++)
        {
            cvLine(regions.raw(), corn[i].lower,corn[i].upper,cvScalar(corn[i].index),15,8,0);
        }
    }
    vector<pixel_accumulator> accum;
    pixel_accumulator temp_px;
    temp_px.count=0;
    accum.assign(19,temp_px);
    for (int y=0; y<input.getHeight(); y++)
    {
        for (int x=0; x<input.getWidth(); x++)
        {
            if (regions[y][x] != 0)
            {
                accum[regions[y][x]].count++;
                accum[regions[y][x]].value+=intensity(input[y][x]);
            }
        }
    }
    for (int i=8; i<corn.size(); i++)
    {
        cvLine(input.raw(), corn[i].lower,corn[i].upper,cvScalar(255,0,0),15,8,0);
    }
    vector<double> in;
    in.assign(18,0.0);
    for (int i=0; i<in.size(); i++)
    {
        if (accum[i+1].value/accum[i+1].count >= 0.0)
        {
            in[i]=accum[i+1].value/accum[i+1].count;
        }
        else
        {
            in[i]=-1.0;
        }
    }
    // 5 levels
    cout.setf(ios::fixed);
    cout.setf(ios::showpoint);
    // this shows the readings of the initial position.
    // the pattern is the same as used in the AI playing field viewer.
    cout << setprecision(2);
    cout << " " << in[1-1] << setw(5*6) << in[16-1] << endl;
    cout << setw(5) << " " << in[4-1] << setw(5*4) << in[13-1] << endl;
    cout << " " << in[2-1] << setw(5*2) << in[7-1] << setw(5*2) << in[11-1] << setw(5*2) << in[17-1] << endl;
    cout << setw(5) << " " << in[5-1] << setw(5*2) << in[9-1] << setw(5*2) << in[14-1] << endl;
    cout << " " << in[3-1] << setw(5*2) << in[8-1] << setw(5*2) << in[12-1] << setw(5*2) << in[18-1] << endl;
    cout << setw(5) << " " << in[6-1] << setw(5*2) << in[10-1] << setw(5*2) << in[15-1] << endl;

    vector<double> temp_in = in;
    sort(temp_in.begin(),temp_in.end(),descending_cmp);
    double darkest_white = temp_in[5];
    // setting all non-visible cylinders to unknown
    for (int i=0; i<18; i++)
    {
        if (in[i] == -1.0)
        {
            cylinders.cylinder[i].color='u';
        }
        else if (in[i] < darkest_white)
        {
            cylinders.cylinder[i].color='b';
        }
        else
        {
            cylinders.cylinder[i].color='w';
        }
    }
    // this mirrors the playing field configuration, so that the AI gets a complete cylinder configuration,
    // and not just half the field.
    for (int i=0; i<18; i++)
    {
        if (cylinders.cylinder[i].color == 'u')
        {
            //0 == 15, 1 == 16, 2 == 17
            //3 == 12, 4 == 13, 5 == 14
            //6 == 10, 7 == 11
            if (i < 3)
            {
                cylinders.cylinder[i].color=cylinders.cylinder[i+15].color;
            }
            else if ( i < 6 )
            {
                cylinders.cylinder[i].color=cylinders.cylinder[i+9].color;
            }
            else if ( i < 8 )
            {
                cylinders.cylinder[i].color=cylinders.cylinder[i+4].color;
            }
            else if ( i < 12)
            {
                cylinders.cylinder[i].color=cylinders.cylinder[i-4].color;
            }
            else if ( i < 15)
            {
                cylinders.cylinder[i].color=cylinders.cylinder[i-9].color;
            }
            else if ( i < 18)
            {
                cylinders.cylinder[i].color=cylinders.cylinder[i-15].color;
            }
        }
    }
    cylinders.counter=getPickupCount();
    for (int i=0; i<18; i++)
    {
        cout << cylinders.cylinder[i].color << " " << flush;
    }
    cout << endl;
    send_cylinder_positions(cylinders);
    //sleep(5); // making system sleep for 5 seconds.
    //sleep(5);
    /*
    ball_positions balls;
    for(int t=0; t<14; t++) {
        balls.tomato[t].y=1;
        balls.tomato[t].x=1;
    }
    for(int a=0; a<12; a++) {
        balls.orange[a].y=100*a;
        balls.orange[a].x=100*a;
    }
    balls.counter=getPickupCount();
    send_ball_positions(balls);
    */
}


// this routine is used to draw cylinder positions in the image based on ground truth
// own position, and position of all the cylinders.
void all_cylinders(RgbImage& left_input, RgbImage& right_input,uint64_t timestamp)
{
    //printf("204: %d\t 334: %d\t 594: %d\n",short_range,long_range,close_range);
    robot = getOurPosInterpolated(timestamp);
    //robot.psi = 0*M_PI/180.0;
    vector< CvPoint > corn;
    corn_populate(corn);
    for (int i=0; i<corn.size(); i++)
    {
        cylinder_draw(left_input, right_input, corn[i].y, corn[i].x);
        //printf("corn: %d y: %d, x: %d\n",i, corn[i].y, corn[i].x);
    }
    /*
    printf("Distance( 1.50): %lf\n",gt.distance_from_vertical_index(270+30));
    printf("Distance( 1.00): %lf\n",gt.distance_from_vertical_index(330+30));
    printf("Distance( 0.50): %lf\n",gt.distance_from_vertical_index(480+30));
    printf("Distance( 0.25): %lf\n",gt.distance_from_vertical_index(687+30));
    */
}


// this routine is used to show map out the enemy robot from the image.
void enemy_mapping(RgbImage& left, RgbImage& right, uint64_t timestamp)
{
    robot = getOurPosInterpolated(timestamp);
    robot_position enemy = getEnemyPosInterpolated(timestamp);
    //enemy.y=1500;
    //enemy.x=2000;
    robot_position left_camera;
    robot_position right_camera;
    camera_positions(robot,left_camera,right_camera);
    int rel_y = enemy.y-left_camera.y;
    int rel_x = enemy.x-left_camera.x;
    double distance = sqrt(rel_y*rel_y+rel_x*rel_x);
    // left:
    CvPoint middle = gt.image_position_from_real_coord(rel_y,rel_x, robot.psi, 0);
    CvPoint u_l,l_r;
    u_l.y=0;
    u_l.x=middle.x-130000/distance;
    l_r.y=middle.y;
    l_r.x=middle.x+130000/distance;
    cout << "ENEMY_MAPPING: " << middle.y << " " << middle.x << "Enemy is at: " << enemy.y << " " << enemy.x << endl;
    cvRectangle(left.raw(),u_l,l_r,CV_RGB(0,0,255),CV_FILLED, 8,0);
    //cvLine(left.raw(),l_r,u_l,cvScalar(255,0,0),50,8,0);
    // right:
    rel_y = enemy.y-right_camera.y;
    rel_x = enemy.x-right_camera.x;
    distance = sqrt(rel_y*rel_y+rel_x*rel_x);

    middle = gt.image_position_from_real_coord(rel_y,rel_x, robot.psi, 0);
    u_l.y=0;
    u_l.x=middle.x-130000/distance;
    l_r.y=middle.y;
    l_r.x=middle.x+130000/distance;
    cvRectangle(right.raw(),u_l,l_r,CV_RGB(0,0,255),CV_FILLED, 8,0);
}


void alter_focal_length(double value)
{
    gt.setFL(gt.getFL()+value);
    printf("Focal Length: %lf\n",gt.getFL());
}

void alter_view_height(double value)
{
    gt.setViewHeight(gt.getViewHeight()+value);
    printf("View Height: %lf\n",gt.getViewHeight());
}

void alter_angle(double value)
{
    gt.setDownwardsAngle(gt.getDownwardsAngle()+value);
    printf("Downwards Angle: %lf\n",gt.getDownwardsAngle());
}

pair<CvPoint,CvPoint> find_border_point(robot_position &left,robot_position &right, int abs_y, int abs_x)
{
    CvPoint p_left={100000,100000};
    CvPoint p_right={100000,100000};
    int rel_y = abs_y-left.y;
    int rel_x = abs_x-left.x;
    //double l_ang=angle_from_position_and_orientation(rel_y, rel_x, left.psi);
    //double r_ang=angle_from_position_and_orientation(abs_y-right.y, abs_x-right.x, right.psi);
    //if(l_ang<90.0 && l_ang>-90.0 && r_ang<90.0 && r_ang>-90.0) {
    p_left=gt.image_position_from_real_coord(rel_y, rel_x, left.psi, 50);
    rel_y = abs_y-right.y;
    rel_x = abs_x-right.x;
    p_right=gt.image_position_from_real_coord(rel_y, rel_x, right.psi, 50);
    //}
    /*
    if(p_left.x < 0 || p_right.x < 0 || p_left.x > 1024 || p_right.x > 1024) {

        p_left.y = 768-((p_left.y)%768);
        p_right.y = 768-((p_right.y)%768);
    }
    */
    //}
    return pair<CvPoint,CvPoint>(p_left,p_right);
}

void check_line(int a, int b, int &highest, int &lowest) {
        if(a > highest) {
            highest = a;
        }
        if(b > highest) {
            highest = b;
        }
        if(a < lowest) {
             lowest = a;
        }
        if(b < lowest) {
            lowest = b;
        }
}

void point_draw(BwImage& left, BwImage& right,pair<CvPoint, CvPoint> &a, pair<CvPoint, CvPoint> &b)
{
    //if(a.first.y != 100000 && a.first.x != 100000 && b.first.y != 100000 && b.first.x != 100000) {
    CvPoint u_l,l_r;
    if (border_is_in_view(a.first) || border_is_in_view(b.first))
    {
        //u_l=a.first;
        //u_l.y=0;
        cvLine(left.raw(),a.first,b.first,cvScalar(255),1,8,0);
        //cvRectangle(left.raw(),u_l,b.first,CV_RGB(0,0,255),CV_FILLED, 8,0);
    }
    if (border_is_in_view(a.second) || border_is_in_view(b.second))
    {
        //u_l=a.second;
        //u_l.y=0;
        //cvRectangle(right.raw(),u_l,b.second,CV_RGB(0,0,255),CV_FILLED, 8,0);
        cvLine(right.raw(),a.second,b.second,cvScalar(255),1,8,0);
    }
    //}
}

void point_draw(RgbImage& left, RgbImage& right,pair<CvPoint, CvPoint> &a, pair<CvPoint, CvPoint> &b)
{
    //if(a.first.y != 100000 && a.first.x != 100000 && b.first.y != 100000 && b.first.x != 100000) {
    CvPoint u_l,l_r;
    if (border_is_in_view(a.first) || border_is_in_view(b.first))
    {
        u_l=a.first;
        u_l.y=0;
        cvLine(left.raw(),a.first,b.first,CV_RGB(0,0,255),1,8,0);
        //cvRectangle(left.raw(),u_l,b.first,CV_RGB(0,0,255),CV_FILLED, 8,0);
    }
    if (border_is_in_view(a.second) || border_is_in_view(b.second))
    {
        u_l=a.second;
        u_l.y=0;
        //cvRectangle(right.raw(),u_l,b.second,CV_RGB(0,0,255),CV_FILLED, 8,0);
        cvLine(right.raw(),a.second,b.second,CV_RGB(0,0,255),1,8,0);
    }
    //}
}

void line_intersect(CvPoint &camera_origin, CvPoint &direction, vector<CvPoint> coordinates)
{
    double x,y;
    int c=0;
    printf("CHECKING\n");
    for (c=0; c<coordinates.size(); c++)
    {
        if (lineSegmentIntersection(camera_origin,direction,coordinates[c], coordinates[(c+1)%4],x, y))
        {
            printf("Line intersect: %d, X: %f, Y: %f (%d,%d)->(%d,%d)\n",c,x,y,camera_origin.y,camera_origin.x,direction.y,direction.x);
        }
    }
}

bool border_is_in_view(CvPoint &p)
{
    if (p.x<0 || p.x>1024 || p.y < 0 || p.y > 768)
    {
        return false;
    }
    else
    {
        return true;
    }
}


bool is_line(double orientation, double min, double max)
{
    if ((orientation*M_PI/180.0) < max-32.0 && (orientation*M_PI/180.0) > 32.0+min)
    {
        return false;
    }
    else
    {
        return true;
    }
}

// this routine draws a border in the image,
// that can be used to map out the parts of the image that are not on the playing field.
void border_draw(frame<RgbPixel>& left, frame<RgbPixel>& right, uint64_t timestamp)
{
    robot = getOurPosInterpolated(timestamp);
    robot_position left_camera;
    robot_position right_camera;
    camera_positions(robot,left_camera,right_camera);

    //CvPoint right_cam_l = rotate_axis_single(right_camera,l_y,l_x);
    //CvPoint right_cam_r = rotate_axis_single(right_camera,l_y,-l_x);
    vector< pair<CvPoint, CvPoint> > border_points;
    pair<CvPoint, CvPoint> temp;
    for (int i=0; i<3000; i+=375)
    {
        temp=find_border_point(left_camera,right_camera,0,i);
        border_points.push_back(temp);
    }
    for (int i=0; i<2104; i+=263)
    {
        temp=find_border_point(left_camera,right_camera,i,3000);
        border_points.push_back(temp);
    }
    for (int i=3000; i>0; i-=375)
    {
        temp=find_border_point(left_camera,right_camera,2104,i);
        border_points.push_back(temp);
    }
    for (int i=2104; i>0; i-=263)
    {
        temp=find_border_point(left_camera,right_camera,i,0);
        border_points.push_back(temp);
    }
    /*
        pair<CvPoint, CvPoint> u_l = find_border_point(left_camera,right_camera,0,0);
        pair<CvPoint, CvPoint> u_ml = find_border_point(left_camera,right_camera,0,750);
        pair<CvPoint, CvPoint> u_m = find_border_point(left_camera,right_camera,0,1500);
        pair<CvPoint, CvPoint> u_mr = find_border_point(left_camera,right_camera,0,2250);
        pair<CvPoint, CvPoint> u_r = find_border_point(left_camera,right_camera,0,3000);
        pair<CvPoint, CvPoint> m_r = find_border_point(left_camera,right_camera,1050,3000);
        pair<CvPoint, CvPoint> l_r = find_border_point(left_camera,right_camera,2100,3000);
        pair<CvPoint, CvPoint> l_m = find_border_point(left_camera,right_camera,2100,1500);
        pair<CvPoint, CvPoint> m_l = find_border_point(left_camera,right_camera,1050,0);
        pair<CvPoint, CvPoint> l_l = find_border_point(left_camera,right_camera,2100,0);
    */
    for (int b=0; b<border_points.size(); b++)
    {
        //point_draw(left.remapped,right.remapped,border_points.at(b),border_points.at((b+1)%border_points.size()));
        point_draw(left.ch1,right.ch1,border_points.at(b),border_points.at((b+1)%border_points.size()));
    }
    /*
        point_draw(left_input,right_input,u_l,u_m);
        point_draw(left_input,right_input,u_m,u_r);
        point_draw(left_input,right_input,u_r,m_r);
        point_draw(left_input,right_input,m_r,l_r);
        point_draw(left_input,right_input,l_r,l_m);
        point_draw(left_input,right_input,l_m,l_l);
        point_draw(left_input,right_input,l_l,m_l);
        point_draw(left_input,right_input,m_l,u_l);
        printf("U_L: (%d,%d), U_R: (%d,%d)\n",u_l.first.y,u_l.first.x,u_r.first.y,u_r.first.x);
        printf("L_L: (%d,%d), L_R: (%d,%d)\n",l_l.first.y,l_l.first.x,l_r.first.y,l_r.first.x);
        */
}

// this draws a point for every cylinder that should be visible in the image.
void cylinder_draw(RgbImage& left_input, RgbImage& right_input, int abs_y, int abs_x)
{
    robot_position left_camera;
    robot_position right_camera;
    camera_positions(robot,left_camera,right_camera);
    // for left camera
    int rel_y = abs_y-left_camera.y;
    int rel_x = abs_x-left_camera.x;
    CvPoint lower = gt.image_position_from_real_coord(rel_y, rel_x, left_camera.psi, 0);
    CvPoint upper = gt.image_position_from_real_coord(rel_y, rel_x, left_camera.psi, 1);
    cvLine(left_input.raw(),lower,upper,cvScalar(255,0,0),5,8,0);
    // for right camera
    rel_y = abs_y-right_camera.y;
    rel_x = abs_x-right_camera.x;
    lower = gt.image_position_from_real_coord(rel_y, rel_x, right_camera.psi, 0);
    upper = gt.image_position_from_real_coord(rel_y, rel_x, right_camera.psi, 1);
    cvLine(right_input.raw(),lower,upper,cvScalar(255,0,0),5,8,0);
}


// this populates the playing field with all the tomatoes (red balls),
// for use as initial configuration for the AI.
void tomato_populate(vector< CvPoint> &tomato)
{
    CvPoint temp;
    temp.y=0;
    temp.x=0;
    tomato.assign(14,temp);
    tomato[0].x=150;
    tomato[0].y=994;

    tomato[12].y = tomato[0].y;
    tomato[12].x=3000-150;

    tomato[1].x = tomato[0].x;
    tomato[1].y=1494;

    tomato[13].y=tomato[1].y;
    tomato[13].x=3000-150;

    tomato[2].x=600;
    tomato[2].y=1244;

    tomato[10].y=tomato[2].y;
    tomato[10].x=3000-600;

    tomato[3].x=tomato[2].x;
    tomato[3].y=1744;

    tomato[11].y=tomato[3].y;
    tomato[11].x=3000-600;

    tomato[4].x=1050;
    tomato[4].y=1494;

    tomato[8].y=tomato[4].y;
    tomato[8].x=3000-1050;

    tomato[5].x=tomato[4].x;
    tomato[5].y=1994;

    tomato[9].y=tomato[5].y;
    tomato[9].x=3000-1050;

    tomato[6].x=1500;
    tomato[6].y=1244;

    tomato[7].x=tomato[6].x;
    tomato[7].y=1744;
}


// this populates the playing field with all the corns (cylinders),
// for use as initial configuration for the AI.
void corn_populate(vector< CvPoint > &corn)
{
    CvPoint start;
    corn.assign(18,start);
    start.x=150;
    start.y=740; // 740 is actual.
    for (int i=0; i<18; i++)
    {
        int n=6;
        if (i<15) n=5;
        if (i<12) n=4;
        if (i<10) n=3;
        if (i<8) n=2;
        if (i<6) n=1;
        if (i<3) n=0;
        corn[i].x=start.x+n*450;
        switch (i)
        {
        case 0:
        case 15:
            corn[i].y=start.y+0*250;
            break;
        case 3:
        case 12:
            corn[i].y=start.y+1*250;
            break;
        case 1:
        case 6:
        case 10:
        case 16:
            corn[i].y=start.y+2*250;
            break;
        case 4:
        case 8:
        case 13:
            corn[i].y=start.y+3*250;
            break;
        case 2:
        case 7:
        case 11:
        case 17:
            corn[i].y=start.y+4*250;
            break;
        case 5:
        case 9:
        case 14:
            corn[i].y=start.y+5*250;
            break;
        default:
            break;
        }
    }
}

// this is a public domain implementation found at http://alienryderflex.com/intersect/
bool lineSegmentIntersection(CvPoint &a, CvPoint &b, CvPoint &c, CvPoint &d,double &x, double &y)
{

    double  distAB, theCos, theSin, newX, ABpos ;

    //  Fail if either line segment is zero-length.
    if (a.x==b.x && a.y==b.y || c.x==d.x && c.y==d.y) return false;

    //  Fail if the segments share an end-point.
    if (a.x==c.x && a.y==c.y || b.x==c.x && b.y==c.y
            ||  a.x==d.x && a.y==d.y || b.x==d.x && b.y==d.y)
    {
        return false;
    }

    //  (1) Translate the system so that point A is on the origin.
    b.x-=a.x;
    b.y-=a.y;
    c.x-=a.x;
    c.y-=a.y;
    d.x-=a.x;
    d.y-=a.y;

    //  Discover the length of segment A-B.
    distAB=sqrt(b.x*b.x+b.y*b.y);

    //  (2) Rotate the system so that point B is on the positive X axis.
    theCos=b.x/distAB;
    theSin=b.y/distAB;
    newX=c.x*theCos+c.y*theSin;
    c.y  =c.y*theCos-c.x*theSin;
    c.x=newX;
    newX=d.x*theCos+d.y*theSin;
    d.y  =d.y*theCos-d.x*theSin;
    d.x=newX;

    //  Fail if segment C-D doesn't cross line A-B.
    if (c.y<0. && d.y<0. || c.y>=0. && d.y>=0.) return false;

    //  (3) Discover the position of the intersection point along line A-B.
    ABpos=d.x+(c.x-d.x)*d.y/(d.y-c.y);

    //  Fail if segment C-D crosses line A-B outside of segment A-B.
    if (ABpos<0. || ABpos>distAB) return false;

    //  (4) Apply the discovered position to line A-B in the original coordinate system.
    x=a.x+ABpos*theCos;
    y=a.y+ABpos*theSin;

    //  Success.
    return true;
}


