#include <sys/time.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
#include <stdint.h>
#include <inttypes.h>

#include <iostream>
#include <cstdlib>

#include <cstdio>
#include "camera.hpp"
#include "individual_processing.hpp"
#include "state.hpp"
#include "position.hpp"

#define MAX_DEVICES 64
#define MAX_FORMATS 64
#define MAX_PROPERTIES 64
#define FOURCC(a,b,c,d) (unsigned int)((((unsigned int)d)<<24)+(((unsigned int)c)<<16)+(((unsigned int)b)<<8)+a)
#define CLIP(a,l,u) ((a)<(l)?(l):(((a)>(u)) ? (u) : (a) ) )

#include "dc1394/dc1394.h"

/* uncomment the following to drop frames to prevent delays */
#define MAX_PORTS   4
#define MAX_CAMERAS 2
#define NUM_BUFFERS 2

#include <ctime>
using namespace std;

unsigned char *cam_left_buffer = NULL;
unsigned char *cam_right_buffer = NULL;

/* declarations for libdc1394 */
uint32_t numCameras = 0;
dc1394camera_t *cameras[MAX_CAMERAS];
dc1394featureset_t features;
dc1394video_frame_t * frames[MAX_CAMERAS];

/* declarations for video1394 */
char *device_name=NULL;

/* Other declarations */
unsigned long frame_length;
long frame_free;
int adaptor=-1;

int connection=-1;

void capture_cleanup(void)
{
    int i;
    for (i=0; i < numCameras; i++)
    {
        dc1394_video_set_transmission(cameras[i], DC1394_OFF);
        dc1394_capture_stop(cameras[i]);
    }
}

int exposure_to_microsec(int exposure)
{
	if(exposure <= 500) {
		return exposure;
	}
	if(exposure <= 1000) {
		return (exposure-500)*10+500;
	}
	if(exposure <= 1705) {
		return (exposure-1000)*100+5500;
	}
	if(exposure <= 2399) {
		return ((exposure-1705)+76)*1000;
	}
	if(exposure <= 2902) {
		return ((exposure-2399)*10+770)*1000;
	}
	if(exposure <= 3304) {
		return ((exposure-2902)*100+5800)*1000;
	}
	if(exposure <= 3508) {
		return ((exposure-3304)*1000+46000)*1000;
	}
	if(exposure <= 3843) {
		return ((exposure-3508)*10+250)*1000000;
	}
}

void* slave_thread(void* args)
{
    while (!stop)
    {
        new_frame(1,cam_right_buffer);
    }
    pthread_exit(0);

}

void* capture_thread(void* args)
{
    int rc;
    double sync_start = get_time();
    cout << "Waiting for cylinder config request: " << flush;
    //int state = STATE_WAITING;
    int count_sync_frames=0;
    uint64_t capture_timestamp=0;
    const int64_t capture_delay=0; // must probably be set to some constant.
    while (!stop)
    {
        if(count_sync_frames % 2 == 0) {
            if (dc1394_capture_dequeue(cameras[1], DC1394_CAPTURE_POLICY_WAIT, &frames[1])!=DC1394_SUCCESS)
                    dc1394_log_error("Failed to capture from camera %d", 1);
            if (dc1394_capture_dequeue(cameras[0], DC1394_CAPTURE_POLICY_WAIT, &frames[0])!=DC1394_SUCCESS)
                   dc1394_log_error("Failed to capture from camera %d", 0);

        } else {
            if (dc1394_capture_dequeue(cameras[0], DC1394_CAPTURE_POLICY_WAIT, &frames[0])!=DC1394_SUCCESS)
                    dc1394_log_error("Failed to capture from camera %d", 0);
            if (dc1394_capture_dequeue(cameras[1], DC1394_CAPTURE_POLICY_WAIT, &frames[1])!=DC1394_SUCCESS)
                   dc1394_log_error("Failed to capture from camera %d", 1);
        }
        memcpy(cam_left_buffer,frames[0]->image,imageSize.width*imageSize.height*2);
        memcpy(cam_right_buffer,frames[1]->image,imageSize.width*imageSize.height*2);

        cam_left.timestamp = frames[0]->timestamp-capture_delay;
        cam_right.timestamp = frames[1]->timestamp-capture_delay;
        for (int camera=0; camera<2; camera++)
        {
            if (frames[camera])
            {
                dc1394_capture_enqueue(cameras[camera], frames[camera]);
            }
        }
        /*
        int cam_done_0=0;
        int cam_done_1=0;
        while(!cam_done_0 || !cam_done_1) {
             if(!cam_done_0) {
                 if(dc1394_capture_dequeue(cameras[0], DC1394_CAPTURE_POLICY_POLL, &frames[0])==DC1394_SUCCESS) {
                     cam_done_0=1;
                 }
             }
             if(!cam_done_1) {
                if(dc1394_capture_dequeue(cameras[1], DC1394_CAPTURE_POLICY_POLL, &frames[1])==DC1394_SUCCESS) {
                     cam_done_1=1;
                 }
             }
        }
        */
        int state = getGlobalState();
        if(state == STATE_RUNNING || state == STATE_CYLINDER_CONFIG) {
            double time_capture = get_time();
            new_frame(0,cam_left_buffer);
            double processing_done = get_time();
            cout << "Time spent processing: " << (processing_done-time_capture)*1000 << " ms" << endl;
        } else if (state == STATE_WAITING) {
            count_sync_frames++;
            double sync_lapsed = get_time();
            cout << "." << flush;
            if(count_sync_frames % 25 == 0) {
                cout << " " << count_sync_frames << "/300" << endl;
                cout << "Waiting for cylinder config request: " << flush;
            }
        }
    }
    printf("stop!\n");
    pthread_exit(NULL);
}

int capture_setup(unsigned long width, unsigned long height, int fps ) {
    int res;
    switch(width) {
        case 1024:
            //res = DC1394_VIDEO_MODE_1024x768_MONO8;
            res = DC1394_VIDEO_MODE_1024x768_YUV422;
            break;
        case 800:
            res = DC1394_VIDEO_MODE_800x600_MONO8;
            break;
        case 640:
            res = DC1394_VIDEO_MODE_640x480_MONO8;
            break;
        default:
            res = DC1394_VIDEO_MODE_1024x768_MONO8;
            break;
    }
    switch(fps) {
        case 7:
            fps = DC1394_FRAMERATE_7_5;
            break;
        case 15:
            fps = DC1394_FRAMERATE_15;
            break;
        case 30:
            fps = DC1394_FRAMERATE_30;
            break;
        default:
            fps = DC1394_FRAMERATE_7_5;
            break;
    }
    imageSize = cvSize(width, height);

    cam_left_buffer = (unsigned char*) malloc(imageSize.width*imageSize.height*2); // 2 bytes = 16-bit
    cam_right_buffer = (unsigned char*) malloc(imageSize.width*imageSize.height*2); // 2 bytes = 16-bit
    int i, j;
    dc1394_t * d;
    dc1394camera_list_t * list;
    //format=XV_UYVY;
    long format=FOURCC('U','Y','V','Y');
    //long format=FOURCC('Y','8','0','0');
    dc1394error_t err;

    d = dc1394_new ();
    if (!d) {
        printf("/dev/dc1394: has wrong permissions or does not exist!\n");
        return -1;
    }
    err=dc1394_camera_enumerate (d, &list);
    DC1394_ERR_RTN(err,"Failed to enumerate cameras");

    if (list->num == 0)
    {
        dc1394_log_error("No cameras found");
        return -1;
    }
    if(list->num == 2) {
        cameras[1] = dc1394_camera_new(d, 582164340023517310); // Bård swapped the circuit boards on the robot, so R = Left, L = Right
        cameras[0] = dc1394_camera_new(d, 582164340023517321);
        if(!cameras[0] || !cameras[1]) {
            dc1394_log_warning("Failed to initialize camera");
            return -1;
        }
    } else {
        dc1394_log_error("Incorrect (%d) cameras found.",list->num);
        return -1;
    }
    /*
    for (i = 0; i < list->num; i++)
    {
        if (j >= MAX_CAMERAS)
            break;
        cameras[j] = dc1394_camera_new (d, list->ids[i].guid);
        cout << "GUID: " << list->ids[i].guid << endl;
        if (!cameras[j])
        {
            dc1394_log_warning("Failed to initialize camera with guid %llx", list->ids[i].guid);
            continue;
        }
        j++;
    }
    */
    numCameras = 2;
    dc1394_camera_free_list (list);

    if (numCameras == 0)
    {
        dc1394_log_error("No cameras found");
        return -1;
    }
    /*
    printf("Resetting bus..\n");
    for(i = 0; i < numCameras; i++) {
        dc1394_reset_bus(cameras[i]);
    }
    */

    /* setup cameras for capture */
    for (i = 0; i < numCameras; i++)
    {

        err=dc1394_video_set_iso_speed(cameras[i], DC1394_ISO_SPEED_400);
        DC1394_ERR_CLN_RTN(err,capture_cleanup(),"Could not set ISO speed");

        err=(dc1394_video_set_mode(cameras[i], (dc1394video_mode_t) res));
        DC1394_ERR_CLN_RTN(err,capture_cleanup(),"Could not set video mode");

        err=dc1394_video_set_framerate(cameras[i], (dc1394framerate_t) fps);
        DC1394_ERR_CLN_RTN(err,capture_cleanup(),"Could not set framerate");

        err=dc1394_capture_setup(cameras[i],NUM_BUFFERS, DC1394_CAPTURE_FLAGS_DEFAULT);
        DC1394_ERR_CLN_RTN(err,capture_cleanup(),"Could not setup camera-\nmake sure that the video mode and framerate are\nsupported by your camera");

    }
    for (i = 0; i < numCameras; i++) {
    // setting features:
    dc1394_feature_set_value(cameras[i],DC1394_FEATURE_BRIGHTNESS,300); // 450
    dc1394_feature_set_value(cameras[i],DC1394_FEATURE_SHARPNESS,250); // 350
    dc1394_feature_set_value(cameras[i],DC1394_FEATURE_GAMMA,16);
    dc1394_feature_set_value(cameras[i],DC1394_FEATURE_SHUTTER,1005); // 950 = 5 ms, 1005 = 6 ms
    dc1394_feature_set_value(cameras[i],DC1394_FEATURE_GAIN,200); // 350
    dc1394_feature_set_value(cameras[i],DC1394_FEATURE_HUE,280);
    dc1394_feature_set_mode(cameras[i],DC1394_FEATURE_WHITE_BALANCE, DC1394_FEATURE_MODE_MANUAL); // for manual white balance control
    dc1394_feature_whitebalance_set_value(cameras[i],511,400); // 400 gives slightly too high green.
    }

    err=dc1394_camera_set_broadcast(cameras[0],DC1394_TRUE);
    DC1394_ERR_CLN_RTN(err,capture_cleanup(),"Could not enable broadcast.");
    err=dc1394_camera_set_broadcast(cameras[1],DC1394_TRUE);
    DC1394_ERR_CLN_RTN(err,capture_cleanup(),"Could not enable broadcast.");
    err=dc1394_video_set_transmission(cameras[0], DC1394_ON);
    DC1394_ERR_CLN_RTN(err,capture_cleanup(),"Could not start camera iso transmission");
    //err=dc1394_camera_set_broadcast(cameras[0],DC1394_FALSE);
    //DC1394_ERR_CLN_RTN(err,cleanup(),"Could not disable broadcast.");

    fflush(stdout);
    if (numCameras < 1)
    {
        perror("no cameras found :(\n");
        capture_cleanup();
        return -1;
    }
}
