#include "canny.hpp"
#include <cmath>
#include <iostream>
#include "wrapper.hpp"

using namespace std;

int compare(const void* a, const void* b) {
	return ( *(int*)a - *(int*)b );
}

Canny::Canny(int x_size, int y_size)
{
	height=y_size;
	width=x_size;
	upper_threshold=60;
	lower_threshold=30;
	vector<double> temp;
	temp.assign(width,0);
	grad_direction.assign(height,temp);
	gradient.assign(height,temp);
	processed_image.assign(height,temp);
}

void Canny::set_threshold(int lower, int upper)
{
	// recommended is upper_threshold = 2*lower_threshold
	lower_threshold=lower;
	upper_threshold=upper;
}

void Canny::perform(IplImage* raw_input, CvMat* dx, CvMat* dy)
{
	BwImage input(raw_input);
	//cout << "calculating gradient.." << endl;
	calculate_gradient(input,dx,dy);
	//cout << "suppressing non-maxima.." << endl;
	suppress_nonmaxima();
	//cout << "detecting edges.." << endl;
	detect_edge();
	//cout << "writing processed image" << endl;
	for(int y=0; y<height; y++)
		for(int x=0; x<width; x++)
		{
			input[y][x] = processed_image[y][x];
		}
}

void Canny::calculate_gradient(BwImage& input,CvMat* raw_dx,CvMat* raw_dy)
{
    MatShort dx(raw_dx);
	MatShort dy(raw_dy);
	for(int y=0; y < height; y++)
		for (int x=0; x<width; x++)
		{
			grad_direction[y][x] = atan2(dx[y][x],dy[y][x]); // angle
			gradient[y][x] = sqrt(dx[y][x]*dx[y][x]+dy[y][x]*dy[y][x]); // gradient strength
		}
}

void Canny::suppress_nonmaxima()
{
	int delx[4] = {1,1,0,-1};
	int dely[4] = {0,1,1,1};
	int x, y, direction;
	double Imax=0;
	double gd=0;
	int count[4]={0,0,0,0};
	for(y=1; y<height-1; y++)
		for(x=1; x<width-1; x++)
		{
			gd = grad_direction[y][x];
			if(gd < 0)
				gd += M_PI;
			direction = (int) round((gd / M_PI) * 4.0) % 4;
			if(direction <= 3)
			{
				count[direction]++;
			} else {
					cerr << "invalid gradient direction" << endl;
			}
			// suppress non-maxima in both directions.
			// (ignores border-cases for now)
			if(gradient[y][x] <= gradient[y+delx[direction]][x+dely[direction]] || gradient[y][x] <= gradient[y-delx[direction]][x-dely[direction]])
			{
					gradient[y][x] = 0;
			}
		}
	for(y=0; y<height; y++)
		for(x=0; x<width; x++)
		{
			if(Imax < gradient[y][x])
				Imax = gradient[y][x];
		}

	if(Imax == 0)
		cerr << "Invalid Imax value!" << endl;

	for(y=0; y<height; y++)
		for(x=0; x<width; x++)
		{
			gradient[y][x] = (int) (255 / Imax * gradient[y][x]);
		}
/*
	cout << "total= " << height*width;
	for(int i=0; i<4; i++)
		cout << " count[" << i << "]= " << count[i];
	cout << endl;
*/
}

void Canny::detect_edge()
{
	int y,x;
	// perform edge detection
	for(y=1; y<height-1; y++)
		for(x=1; x<width-1; x++)
		{
			if(processed_image[y][x] == 0 && gradient[y][x] > upper_threshold)
				follow_edge(y,x);
		}
}

void Canny::follow_edge(int y, int x)
{
	processed_image[y][x] = 255;
	// check 8-neighbours :)
	for(int a=-1; a<2; a++)
		for(int b=-1; b<2; b++)
			if(gradient[y+a][x+b] > lower_threshold && processed_image[y+a][x+b] == 0)
			{
				if(y+a<height-1 && x+b<width-1 && 0<y+a && 0<x+b)
				{
					follow_edge(y+a,x+b);
				}
			}
}
