#include "communication.hpp"
#include "state.hpp"
#include "main.hpp"
#include "position.hpp"
#include <string>
#include <sstream>
#include <cstdlib>
#include <cstring>
using namespace std;

#define PM_DUMMY 4444

int our_color=-1;

// Global data structures
MessageQueue cvsysQueue(PMQ_CVSYS);
MessageQueue navsysQueue(PMQ_NAVSYS);
MessageQueue aiQueue(PMQ_AI);

//int object_pickup_counter=0;
#define ALT_OK 1

// recieving messages.


// AI
void statusCheck()
{
    cvsys_systemcheck result;
    result.camera_status = ALT_OK;
    aiQueue.sendStruct(PM_AI_CVSYS_SYSTEMCHECK, PRIORITY_VERY_HIGH, sizeof(result), (char*) &result);
}

void register_object_pickup(posix_message_t &pm) {
    increasePickupCount();
}

void cylinder_configuration(posix_message_t &pm)
{
    cout << "Cylinder configuration request received!" << endl;
    robot_position request;
    memcpy((void *)(&request),(void *) pm.data, sizeof(robot_position));
    printf("Position: Y: %d, X: %d, Angle: %f\n",request.y, request.x, request.psi);
    //request.timestamp.tv_sec=12000000;
    //request.timestamp.tv_usec=0;
    //OurPositionAdd(request);
    //OurPositionAdd(request);
    //OurPositionAdd(request);
    if(getGlobalState() == STATE_WAITING) {
        if(request.x < 1500) {
            our_color=0;
        } else {
            our_color=1;
        }
        setGlobalState(STATE_CYLINDER_CONFIG);
    }
}

// NAVSYS
void Navsys_ourPosition(posix_message_t &pm)
{
    robot_position data;
    memcpy((void *)(&data),(void *) pm.data, sizeof(data));
    OurPositionAdd(data);
}

void Navsys_EnemyPosition(posix_message_t &pm)
{
    robot_position data;
    memcpy((void *)(&data),(void *) pm.data, sizeof(data));
    EnemyPositionAdd(data);
}


void* communication_thread(void *args)
{
    cvsysQueue.clear();
    aiQueue.clear();
    //setGlobalState(STATE_WAITING);
    posix_message_t pm;
    while (!stop)
    {
        cvsysQueue.receive(&pm);
        switch (pm.id)
        {
        case PM_CVSYS_OBJECT_PICKUP:
            register_object_pickup(pm);
            break;
        case PM_CVSYS_GET_CYLINDER_CONFIG:
            cylinder_configuration(pm);
            break;
        case PM_CVSYS_ROBOT_POSITION:
            Navsys_ourPosition(pm);
            break;
        case PM_CVSYS_OPPONENT_POSITION:
            Navsys_EnemyPosition(pm);
            break;
        case PM_SYSTEMCHECK:
            statusCheck();
            break;
        case PM_CVSYS_STOP:
            pthread_exit(0);
            break;
        default:
            break;
        }
    }
}

// sending messages

void send_stop_to_self() {
    cvsysQueue.sendCommand(PM_CVSYS_STOP, PRIORITY_LOW);
}

void send_cylinder_positions(cylinder_config &cylinders)
{
    //aiQueue.sendStruct(PM_AI_CYLINDER_CONFIG, PRIORITY_HIGH, sizeof(cylinder_config), (char*) &cylinders);
}

void send_ball_positions(ball_positions &balls) {
    //aiQueue.sendStruct(PM_AI_BALL_POSITIONS, PRIORITY_HIGH, sizeof(ball_positions), (char *) &balls);
}


/*
void send_position_correction(coord &position, double orientation) {
    pos result;
    result.robot_position = position;
    result.robot_orientation = orientation;
    navsysQueue.sendStruct(PM_NAVSYS_POSITION_CORRECTION, PRIORITY_HIGH, (int) sizeof(result), (char *) &result);
}

*/
/*
int main()
{
    cvsysQueue.clear();
    cout << endl;
    //aiQueue.clear();
    cout << endl;
    cout << "Cvsys Communication started!" << endl;
    communication_thread(NULL);
}
*/

