#include "conversion.hpp"
#include <numeric>
#include <iostream>

using namespace std;

IplImage *exclusion;
vector<unsigned char> intensity_exclusion;
opt_struct options;

void color_convert(unsigned char *input, IplImage *raw_output, BwImage& intensity, CvSize& imageSize) {
    /*
	switch(options.color) {
		case 0:
			YUYVtoBW(input, raw_output, imageSize);
			break;
		case 1:
			YUYVtoYUV(input, raw_output, imageSize);
			break;
		case 2:
            YUYVtoBW(input, intensity.raw(), imageSize);
			YUYVtoRGB(input, raw_output, imageSize);
			break;
	    case 3:
            YUYVtoYUV(input, raw_output, imageSize);
            break;
		default:
			break;
	}
	*/
	RgbImage output(raw_output);
	UYVYtoRGB(input,output);
	//Y800toRGB(input,output);
}

// converts 8-bit RAW to BWImage
void Y800toBW(unsigned char *input, BwImage& output)
{
    for (int y=0; y<output.getHeight(); y++)
    {
#pragma unroll(4)
        for (int x=0; x<output.getWidth(); x ++)
        {
            output[y][x]=input[y*output.getWidth()+x];
        }
    }
}

void BuffertoRGB(unsigned char* input,RgbImage& output)
{
    memcpy(output.raw()->imageData,input,output.getWidth()*output.getHeight()*3);
}

// converts 8-bit Bayer pattern to RGBImage
// it uses the bayer decoding routines in libdc1394, and the method can be selected by changing
// DC1394_BAYER_METHOD_SIMPLE to various others, such as DC1394_BAYER_METHOD_BILINEAR etc.
void Y800toRGB(unsigned char *input, RgbImage& output) {
        unsigned char* temp_buff = (unsigned char*) malloc(output.getWidth()*output.getHeight()*3);
        dc1394_bayer_decoding_8bit(input,temp_buff,output.getWidth(),output.getHeight(),DC1394_COLOR_FILTER_GRBG,DC1394_BAYER_METHOD_SIMPLE);
        BuffertoRGB(temp_buff,output);
        free(temp_buff);
}

// this is the routine required to turn the YUV format on the UniBrain Fire-i cameras into RGB.
void UYVYtoRGB(unsigned char *input, RgbImage& output)
{
    unsigned char y1, u, v, y2;
    for (int y=0; y<output.getHeight(); y++)
    {
#pragma unroll(4)
        for (int x=0; x<output.getWidth(); x +=2)
        {
            u = input[y*output.getWidth()*2+x*2];
            y1 = input[y*output.getWidth()*2+x*2+1];
            v = input[y*output.getWidth()*2+x*2+2];
            y2 = input[y*output.getWidth()*2+x*2+3];
            YUV444toRGB888(output[y][x],y1,u,v);
            YUV444toRGB888(output[y][x+1],y2,u,v);
        }
    }
}


//writes R,G,B components to output pixel
void setRGB(RgbPixel& output, unsigned char &r, unsigned char &g, unsigned char &b)
{
	output.r=r;
	output.g=g;
	output.b=b;
}


void setRGBfromYUV(YuvImage &yuv, RgbImage &rgb, int y, int x)
{
    rgb[y][x].g=yuv[y][x].y;
    rgb[y][x].b=yuv[y][x].u;
    rgb[y][x].r=yuv[y][x].v;
}

int pixel_maximum(RgbPixel &p) {
	if(p.r>=p.g && p.r>=p.b) { return p.r; }
	if(p.g>=p.r && p.g>=p.b) { return p.g; }
	if(p.b>=p.r && p.b>=p.g) { return p.b; }
}

int pixel_minimum(RgbPixel &p) {
	if(p.r<=p.g && p.r<=p.b) { return p.r; }
	if(p.g<=p.r && p.g<=p.b) { return p.g; }
	if(p.b<=p.r && p.b<=p.g) { return p.b; }
}

int HuefromRGB(RgbPixel &p) {
	int max=pixel_maximum(p);
	int min=pixel_minimum(p);
	if(max==min) {
		return -1;
	}
	if(p.r==max) {
		return (60 * (p.g-p.b)/(max-min) + 360)%360;
	}
	if(p.g==max) {
		return 60 * (p.b-p.r)/(max-min) + 120;
	}
	if(p.b==max) {
		return 60 * (p.r-p.g)/(max-min) + 240;
	}
	return -1;
}

int color_val_RGB(RgbPixel &p) {
	int max=pixel_maximum(p);
	int min=pixel_minimum(p);
	if(max==min) {
		return 0;
	}
	if(p.r==max) {
		return 60 * (p.g-p.b)/(max-min) + 360;
	}
	if(p.g==max) {
		return 60 * (p.b-p.r)/(max-min) + 120;
	}
	if(p.b==max) {
		return 60 * (p.r-p.g)/(max-min) + 240;
	}
	return -1;
}

double SatfromRGB(RgbPixel &p) {
	int max=pixel_maximum(p);
	int min=pixel_minimum(p);
	if(max==0) { return 0; }
	return (max-min)/(double) max;
}

void RGB_threshold_on_HSV(IplImage *raw_input, IplImage *raw_output, CvSize& imageSize) {
	RgbImage input(raw_input);
	RgbImage output(raw_output);
	int hueval=0;
	double satval=0;
	for(int y=0; y<imageSize.height; y++) {
		for(int x=0; x<imageSize.width; x++) {
			hueval=color_val_RGB(input[y][x]);
			satval=SatfromRGB(input[y][x]);
/*
			output[y][x].r=((hueval+180)%360)*255/360;
			output[y][x].g=((hueval+180)%360)*255/360;
			output[y][x].b=((hueval+180)%360)*255/360;
*/
/*

			if(hueval>=300) {
				output[y][x].r=128+(hueval-300);
				output[y][x].g=0;
				output[y][x].b=0;
			} else if(hueval>=180) {
				output[y][x].b=128+(hueval-180);
				output[y][x].r=0;
				output[y][x].g=0;
			} else if(hueval>0) {
				output[y][x].g=128+(hueval-60);
				output[y][x].r=0;
				output[y][x].b=0;
			}
*/
/*
			if(hueval>=300 || hueval<=60) {
				output[y][x].r=input[y]
			output[y][x].g=hueval*255/360;
			output[y][x].b=hueval*255/360;
*/
			if(satval<0.40 && pixel_maximum(input[y][x])>240) {
				//cout << satval << endl;
				output[y][x].b=255;
				output[y][x].r=0;
				output[y][x].g=0;
				//output[y][x].g=0;
				//output[y][x].b=0;
			} else {
				output[y][x]=input[y][x];
				//output[y][x].g=0;
				//output[y][x].b=0;
			}
		}
	}
}

/*
vec point(int x, int y) {
    vec ret;
    ret.x=x;
    ret.y=y;
    return ret;
}

double cross_product(vec &a, vec &b) {
	return a.x*b.y-a.y*b.x;
}

bool side(vec &start, vec &end, vec &point) {
    vec val1;
    val1.x=start.x-end.x;
    val1.y=start.y-end.y;
    vec val2;
    val2.x=start.x-point.x;
    val2.y=start.y-point.y;
    if(cross_product(val1, val2)>=0) {
	return true;
    } else {
	return false;
    }
}

bool point_in_triangle(vec p,vec a, vec b,vec c) {
    if(side(a,b,p) && side(b,c,p) && side(c,a,p)) {
		return true;
	} else {
		return false;
	}
}
*/

// writes Y,U,V components to output pixel
void setYUV(YuvPixel& output, unsigned char &y, unsigned char &u, unsigned char &v)
{
    if(options.color==3) {
	RgbImage color_match(exclusion);
	// sets greyscale if component values match exclusion criterias.
	if(intensity_exclusion[y]==0) {
	switch(color_match[u][v].r) {
	    case 255:
            output.y=y;
	        output.v=255;
	        output.u=0;
	        break;
        case 200:
            output.y=y;
	        output.u=255;
	        output.v=0;
	        break;
        case 1:
            output.y=y;
            output.u=0;
            output.v=0;
            break;
        case 0:
            output.y=y;
            output.u=0;
            output.v=255;
            break;
        default:
            output.y=0;
            output.u=0;
            output.v=0;
            break;
	}
	} else {
		output.y=y;
		output.u=0;
		output.v=0;
	}
    } else {
        output.y=y;
        output.u=u;
        output.v=v;
    }
}


// converts YUV 4:4:4 -> RGB 8:8:8 for a single pixel
void YUV444toRGB888(RgbPixel& output, const unsigned char &y, const unsigned char &u, const unsigned char &v)
{
	// FAST NTSC (a bit dark, but by far the fastest)
    //output.r = CLIP((298*(y-16) + 409*(v-128) + 128)>>8,0,255);
    //output.g = CLIP((298*(y-16) - 100*(u-128) - 208*(v-128) + 128)>>8,0,255);
    //output.b = CLIP((298*(y-16) + 516*(u-128) + 128)>>8,0,255);
	// JFIF
	output.r = CLIP(y + 1.402*(v-128),0,255);
	output.g = CLIP(y - 0.34414*(u-128) - 0.71414*(v-128),0,255);
	output.b = CLIP(y + 1.772*(u-128),0,255);
}

// converts a YUYV raw image to a OpenCV Black&White image.
void YUYVtoBW(unsigned char *input, IplImage* raw_output, CvSize& imageSize)
{
	BwImage output(raw_output);
	for(int y=0; y<imageSize.height; y++) {
		for(int x=0; x<imageSize.width; x++) {
			output[y][x]=input[y*imageSize.width*2+x*2];
		}
	}
}

void YUVtoBW(IplImage* raw_input, IplImage* raw_output, CvSize& imageSize)
{
	YuvImage input(raw_input);
	BwImage output(raw_output);
	for(int y=0; y<imageSize.height; y++) {
		for(int x=0; x<imageSize.width; x++) {
			output[y][x]=input[y][x].y;
		}
	}
}

void split_3channels(RgbImage& input, BwImage& ch1, BwImage& ch2, BwImage& ch3)
{
    for(int y=0; y<input.getHeight(); y++) {
		for(int x=0; x<input.getWidth(); x++) {
			ch1[y][x]=input[y][x].r;
			ch2[y][x]=input[y][x].g;
			ch3[y][x]=input[y][x].b;
		}
	}
}

// converts a YUYV raw image to a OpenCV RGB image.
void YUYVtoRGB(unsigned char *input, IplImage* raw_output,CvSize& imageSize)
{
	RgbImage output(raw_output);
    RgbImage color_match(exclusion);
	unsigned char y1, u, v, y2;
	for(int y=0; y<imageSize.height; y++) {
		#pragma unroll(4)
		for(int x=0; x<imageSize.width; x +=2) {
			// YUV422 to RGB
			// input (u,y1,v,y2), output(R,G,B,R,G,B)
			y1 = input[y*imageSize.width*2+x*2];
			u = input[y*imageSize.width*2+x*2+1];
			y2 = input[y*imageSize.width*2+x*2+2];
			v = input[y*imageSize.width*2+x*2+3];
			YUV444toRGB888(output[y][x],y1,u,v);
			YUV444toRGB888(output[y][x+1],y2,u,v);
		}
	}
}

// converts a YUYV raw image to a OpenCV YUV image.
void YUYVtoYUV(unsigned char *input, IplImage* raw_output, CvSize& imageSize)
{
	YuvImage output(raw_output);
	unsigned char y1, u, v, y2;
	for(int y=0; y<imageSize.height; y++) {
		#pragma unroll(4)
		for(int x=0; x<imageSize.width; x += 2) {
			y1 = input[y*imageSize.width*2+x*2];
			u = input[y*imageSize.width*2+x*2+1];
			y2 = input[y*imageSize.width*2+x*2+2];
			v = input[y*imageSize.width*2+x*2+3];
			setYUV(output[y][x],y1,u,v);
			setYUV(output[y][x+1],y2,u,v);
		}
	}
}

// converts a YUYV raw image to a OpenCV Black&White image.
void YUYVtoBW_matched(unsigned char *input, IplImage* raw_output, CvSize& imageSize)
{
	BwImage output(raw_output);
        unsigned char y1, u, v, y2;
        for(int y=0; y<imageSize.height; y++) {
                for(int x=0; x<imageSize.width; x+=2) {
                        y1 = input[y*imageSize.width*2+x*2];
                        u = input[y*imageSize.width*2+x*2+1];
                        y2 = input[y*imageSize.width*2+x*2+2];
                        v = input[y*imageSize.width*2+x*2+3];
			if(v>=165) {
				output[y][x]=y1;
				output[y][x+1]=y2;
			} else {
				output[y][x]=0;
				output[y][x+1]=0;
			}
                }
        }
}

