#include "correspondence.hpp"
using namespace std;

// simple block-matching routine for correspondence, using SAD.
double block_match(BwImage& left, BwImage& right, int y0, int x0_r, int x0_l, int y_size, int x_size)
{
    double sum_left=0.0,sum_right=0.0, sum=0.0;
    double mse=0.0;
    double n=y_size*x_size;
    double left_avg=0.0,right_avg=0.0;
    for (int y=y0; y<y0+y_size; y++)
    {
        for (int i=0; i<x_size; i++)
        {
            sum_right += right[y][x0_r+i];
            sum_left += left[y][x0_l+i];
        }
    }
    left_avg=sum_left/n;
    right_avg=sum_right/n;
    for (int y=y0; y<y0+y_size; y++)
    {
        for (int i=0; i<x_size; i++)
        {
            sum += abs((right[y][x0_r+i]-right_avg)-(left[y][x0_l+i]-left_avg));
        }
    }
    //mse=pow((sum/(y_size*x_size)),2.0);
    return sum/(y_size*x_size);
}

// this calculates a 2D histogram using chromatic red and chromatic blue as components.
// y-axis: chromatic red
// x-axis: chromatic blue
// it also calculates the distribution of the intensities.
void calculate_histogram(IplImage* raw_input, IplImage* raw_output)
{
    YuvImage input(raw_input);
    RgbImage output(raw_output);
    vector< vector<unsigned int> > rg_histogram;
    vector<unsigned int> intensity_histogram;
    intensity_histogram.assign(256,0);
    rg_histogram.assign(256,intensity_histogram);
    // y is chromatic blue
    // x is chromatic red
    int max_val=1,max_intensity=1,val;
    RgbImage color_match(exclusion);
    for (int y=0; y<imageSize.height; y++)
    {
        for (int x=0; x<imageSize.width; x++)
        {
            rg_histogram[input[y][x].u][input[y][x].v]++;
            if (input[y][x].u != input[y][x].v) intensity_histogram[input[y][x].y]++;
            if (rg_histogram[input[y][x].u][input[y][x].v] > max_val)
            {
                if (input[y][x].u != 0 && input[y][x].v != 0)
                {
                    if (input[y][x].u != input[y][x].v)
                    {
                        max_val = rg_histogram[input[y][x].u][input[y][x].v];
                    }
                }
            }
            if (intensity_histogram[input[y][x].y] > max_intensity)
            {
                if (input[y][x].u != input[y][x].v)
                {
                    max_intensity = intensity_histogram[input[y][x].y];
                }
            }
        }
    }
    for (int r=0; r<256; r++)
    {
        for (int g=0; g<256; g++)
        {
            val= rg_histogram[r][g]*255/max_val;
            if (val<32)
            {
                output[r][g].r=rg_histogram[r][g]*255*8/max_val;
            }
            else
            {
                output[r][g].g =val;
            }
            switch(color_match[r][g].r) {
                case 255:
                    output[r][g].r=255;
                    output[r][g].g=0;
                    output[r][g].b=0;
                    break;
                case 200:
                    output[r][g].r=0;
                    output[r][g].g=255;
                    output[r][g].b=0;
                    break;
                case 1:
                    output[r][g].r=g;
                    output[r][g].b=r;
                    output[r][g].g=128;
                    break;
                case 0:
                    break;
                default:
                    break;
            }
        }
    }
    for (int v=0; v<256; v++)
    {
        if (intensity_exclusion[v]!=255)
        {
            val = intensity_histogram[v]*255/max_intensity;
            block_draw(output, 255, v, 64, 1, val);
        }
        else
        {
            val = -1;
            block_draw(output, 255, v, 64, 1, val);
        }
    }
}

// this draws uniform block at a certain position, with a certain size.
void block_draw(RgbImage& output, int y0, int x0, int y_size, int x_size, int distance)
{
    int peak=0;
    int bottom=65536;
    //cout << "block_draw " << y0 << " " << x0 << " " << disp << endl;
    for (int y=y0; y<y0+y_size; y++)
    {
        for (int x=x0; x<x0+x_size; x++)
        {
            //pixel_color(output[y][x], distance, global_rangesize, global_ranges, peak,bottom);
            if (distance==-1)
            {
                output[y][x].r=255;
            }
            else
            {
                output[y][x].r=CLIP(distance,0,255);
                output[y][x].g=0;
                output[y][x].b=0;
            }
            //output[y][x].b=CLIP(disp+128,0,255);
        }
    }
}

// this tries to find corresponding circles in the left and right images.
void hough_matching_correspondence(CvSeq* left_circles, CvSeq* right_circles, IplImage* raw_output, IplImage* location)
{
    // search circles for matches.
    cvSet(location,cvScalar(0,0,0));
    CvPoint ball;
    int matches=0;
    for (int i=0; i < left_circles->total; i++)
    {
        float *left_p = (float *) cvGetSeqElem( left_circles, i);
	cout << "left: " << left_p[0] << " " << left_p[1] << " " << left_p[2] << "(out of " << left_circles->total << ")" << endl;
        for (int j=0; j< right_circles->total; j++)
        {
            float *right_p = (float *) cvGetSeqElem( right_circles, j);
	    cout << "right: " << right_p[0] << " " << right_p[1] << " " << right_p[2] << "(out of " << right_circles->total << ")" << endl;
            // checking if the points are on reasonably the same line.
            cout << "DEBUG: "<< i << "," << j << " y-diff: " << abs(left_p[1]-right_p[1]) << " size: " << abs(left_p[2]-right_p[2]) << endl;
            if (abs(left_p[1]-right_p[1])<10.0)
            {
                // and of the ~same size
                if (abs(left_p[2]-right_p[2])<10.0)
                {
                    double disp = abs(left_p[0]-right_p[0]);
                    cout << "left: " << left_p[0] << " right: " << right_p[0] << " size: " << left_p[2] << " " << right_p[2]<< endl;
                    //cout << "height error: " << abs(left_p[1]-right_p[1]) << endl;
                    double distance=disparity_to_distance(disp);
                    double distance_over=disparity_to_distance(disp+1)-distance;
                    double distance_under=disparity_to_distance(disp-1)-distance;
                    double x_value = (right_p[0]+left_p[0])/2.0+20;
                    double y_value = (left_p[1]+right_p[1])/2.0+1;
                    cvCircle( raw_output, cvPoint(cvRound(x_value),cvRound(left_p[1])), cvRound(left_p[2]), CV_RGB(255,0,0), 3, 8, 0);
                    cvCircle( raw_output, cvPoint(cvRound(x_value),cvRound(right_p[1])), cvRound(right_p[2]), CV_RGB(0,255,0), 3, 8, 0);
                    //cout << "disparity: " << disp << endl;
                    //cout << "possible position is then: " << endl;
                    ball.x = 400-2.0*(x_value-(imageSize.width/2.0))*2.0*distance_to_view_x(distance)/imageSize.width;
                    ball.y = 600-3.0*distance;
                    cvCircle( location, ball, 5, CV_RGB(0,255,0), 1 );

                    cout << "x: " << (x_value-(imageSize.width/2.0))*2.0*distance_to_view_x(distance)/imageSize.width
                    << " +/-" << distance_to_view_x(distance)/imageSize.width << endl;
                    cout << "y: " << (y_value-(imageSize.height/2.0))*2.0*distance_to_view_y(distance)/imageSize.height
                    <<  " +/-" << distance_to_view_y(distance)/imageSize.height << endl;
                    cout << "z: "<< distance << " " << distance_over << "/" << distance_under << endl;
                    matches++;
                }
            }
        }
    }
}

/*
void edge_matching_correspondence()
{
           int min_disp=640, max_disp=0, x_value=0,y_value=0;
                //cvFindStereoCorrespondenceBM( left_output, right_output, disp,BMState);
                //cvNormalize( disp, vdisp, 0, 16384, CV_MINMAX );
                //value_to_color(disp,raw_output,24000,33000);
		for(int y=0; y<imageSize.height; y++) {
		// for each epipolar line in each image:
			vector<int> left_lines;
			vector<int> right_lines;
			for(int x_r=0; x_r<imageSize.width; x_r++) {
				if(right_output_data[y][x_r].y == 255) {
					right_lines.push_back(x_r);
					output[y][x_r].r=255;
				}
				if(left_output_data[y][x_r].y == 255) {
					output[y][x_r].g=255;
				}
			}
			for(int i=0; i<right_lines.size(); i++) {
				for(int x_l=right_lines[i]; x_l<imageSize.width; x_l++)
				{
					if(left_output_data[y][x_l].y == 255) {
						left_lines.push_back(x_l);
						break;
					}
				}
			}
			//cout << left_lines.size() << " " << right_lines.size() << endl;
			for(int i=0; i+i<right_lines.size() && i<left_lines.size(); i++)
			{
				int disp=abs(left_lines[i]-right_lines[i]);
				int distance=(int) disparity_to_distance(disp);
				if(disp > max_disp) { x_value=right_lines[i]+disp/2.0; }
				y_value=y;
				for(int x=right_lines[i]; x<right_lines[i+1]; x++) {
					disparities[disp]++;
					pixel_color(output[y][x+disp/2], distance, global_rangesize, global_ranges, peak,bottom);
					//output[y][x].g = CLIP(disp,0,255);
				}
			}
		}

		int max=0;
		for(int i=1; i<disparities.size();i++) {
		if(max<disparities[i]) {
			max=disparities[i];
			max_disp=i;
		}
		}
		double distance=disparity_to_distance(max_disp);
		double distance_over=disparity_to_distance(max_disp+1)-distance;
		double distance_under=disparity_to_distance(max_disp-1)-distance;
		cout << "disparity: " << max_disp << endl;
		cout << "possible position is then: " << endl;
		cout << "x: " << (x_value-(imageSize.width/2.0))*2.0/distance_to_view_x(distance) << " +/-" << distance_to_view_x(distance)/imageSize.width << endl;
		cout << "y: " << (y_value-(imageSize.height/2.0))*2.0/distance_to_view_y(distance) <<  " +/-" << distance_to_view_y(distance)/imageSize.height << endl;
		cout << "z: "<< distance << " " << distance_over << "/" << distance_under << endl;
}
*/

unsigned short color_difference(RgbPixel& a, RgbPixel& b)
{
    //return sqrt((a.r-b.r)*(a.r-b.r)+(a.g-b.g)*(a.g-b.g)+(a.b-b.b)*(a.b-b.b));
    //return abs(calc_I1(a)-calc_I1(b))+abs(calc_I2(a)-calc_I2(b))+abs(calc_I3(a)-calc_I3(b));
    return abs(a.r-b.r)+abs(a.g-b.g)+abs(a.b-b.b);
}

double color_difference_cie76(RgbPixel& a, RgbPixel& b)
{
    return sqrt((a.r-b.r)*(a.r-b.r)+(a.g-b.g)*(a.g-b.g)+(a.b-b.b)*(a.b-b.b));
}

// uses luminence for edge direction and color difference for gradient strength.
// will most likely be very slow.
// direction: 0 is vertical, 1 is horizontal.
// luminance can be taken from somewhere else, but for testing purposes, just calculating it is faster.

unsigned char luminance(RgbPixel &pixel)
{
    return (pixel.r*299 + pixel.g*587 + pixel.b*114)/1000;
}

int edge_direction(IplImage* raw_input, int y, int x)
{
    for(int i=-1; i<2; i++) {
        for(int j=-1; j<2; j++) {


        }
    }
}

void RGB_gradient_ignore_black(RgbImage& input, CvMat* raw_output)
{
    MatShort output(raw_output);
    unsigned short grad_x=0,grad_y=0;
    for(int y=1; y<input.getHeight()-1; y++) {
        for(int x=1; x<input.getWidth()-1; x++) {
            if(pixel_is_zero(input[y][x])) {
                grad_x = color_difference(input[y-1][x-1],input[y-1][x+1]);
                grad_x += 2*color_difference(input[y][x-1],input[y][x+1]);
                grad_x += color_difference(input[y+1][x-1],input[y+1][x+1]);
                grad_y = color_difference(input[y+1][x-1],input[y-1][x-1]);
                grad_y += 2*color_difference(input[y+1][x],input[y-1][x]);
                grad_y += color_difference(input[y+1][x+1],input[y-1][x+1]);
                output[y][x] = sqrt(grad_x*grad_x+grad_y*grad_y);
            }
            //output[y][x] = sqrt(grad_x*grad_x+grad_y*grad_y);
        }
    }
}


void RGB_Gradient(IplImage* raw_input, CvMat* raw_output, int direction)
{
    RgbImage input(raw_input);
    //BwImageShort output(raw_output);
    MatShort output(raw_output);
    if(direction==0) {
        for(int y=1; y<imageSize.height-1; y++) {
            for(int x=1; x<imageSize.width-1; x++) {
                        //output[y][x] = color_difference(input[y-1][x-1],input[y-1][x+1]);
                        //output[y][x] += 2*color_difference(input[y][x-1],input[y][x+1]);
                        //output[y][x] += color_difference(input[y+1][x-1],input[y+1][x+1]);
                        //cout << output[y][x] << endl;
            }
        }
    } else if(direction==1) {
        for(int y=1; y<imageSize.height-1; y++) {
            for(int x=1; x<imageSize.width-1; x++) {
                        //output[y][x] = color_difference(input[y+1][x-1],input[y-1][x-1]);
                        //output[y][x] += 2*color_difference(input[y+1][x],input[y-1][x]);
                        //output[y][x] += color_difference(input[y+1][x+1],input[y-1][x+1]);
            }
        }
    } else {
        unsigned short grad_x=0,grad_y=0;
        for(int y=1; y<imageSize.height-1; y++) {
            for(int x=1; x<imageSize.width-1; x++) {
                        grad_x = color_difference(input[y-1][x-1],input[y-1][x+1]);
                        grad_x += 2*color_difference(input[y][x-1],input[y][x+1]);
                        grad_x += color_difference(input[y+1][x-1],input[y+1][x+1]);
                        grad_y = color_difference(input[y+1][x-1],input[y-1][x-1]);
                        grad_y += 2*color_difference(input[y+1][x],input[y-1][x]);
                        grad_y += color_difference(input[y+1][x+1],input[y-1][x+1]);
                        output[y][x] = sqrt(grad_x*grad_x+grad_y*grad_y);
                        //output[y][x] = sqrt(grad_x*grad_x+grad_y*grad_y);
            }
        }
    }
}

void gradient_merge(CvMat* raw_dx, CvMat* raw_dy, CvMat *raw_output) {
    MatShort dx(raw_dx);
    MatShort dy(raw_dy);
    MatShort output(raw_output);
    for(int y=0; y<dy.getHeight(); y++)
    {
        for(int x=0; x<dy.getWidth(); x++)
        {
            output[y][x] = sqrt(dx[y][x]*dx[y][x]+dy[y][x]*dy[y][x]);
        }
    }
}

// this function takes in an image, and its corresponding edge image,
// and calculates the edge directions and displays them.
void edge_data(IplImage* raw_input, IplImage* raw_output)
{
    RgbImage output(raw_output);
    BwImage edge_img(cam_left.edges);
    // using Sobel-operator to get gradient in x and y direction.
    for (int y=0; y<imageSize.height; y++)
    {
        //const uchar* edges_row = edges->data.ptr + y*edges->step;
        const short* dx_row = (const short*)(cam_left.dx->data.ptr + y*cam_left.dx->step);
        const short* dy_row = (const short*)(cam_left.dy->data.ptr + y*cam_left.dy->step);
        for (int x=0; x<imageSize.width; x++)
        {
            if (edge_img[y][x]!=0)
            {
                float vx, vy, sx, sy;
                vx = dx_row[x];
                vy = dy_row[x];
                float mag = sqrt(vx*vx+vy*vy);
                if (mag < 1)
                {
                    continue;
                }
                //assert( mag >= 1 );
                sx = cvRound(vx/mag);
                sy = cvRound(vy/mag);
                if (sx == 1)
                {
                    output[y][x].r=255;
                }
                else if ( sx == 0)
                {
                    output[y][x].r=255;
                    output[y][x].b=255;
                }
                else if ( sx == -1)
                {
                    output[y][x].b=255;
                }
                if (sy == 1)
                {
                    output[y][x].g=255;
                }
                else if (sy == 0)
                {
                    output[y][x].g=128;
                }
                else if (sy == -1)
                {
                    output[y][x].g=0;
                }
                //output[y][x].g=edges_row[x];
                //output[y][x].b=(sy*128 + 127);
                //cout << "sx: " << sx << " sy: " << sy << endl;
            }
        }
    }
    //cvReleaseMat(&dx);
    //cvReleaseMat(&dy);
    //cvReleaseImage(&edges);
}

void edge_data_verbose(frame<RgbPixel> &camera, IplImage* raw_output)
{
    RgbImage output(raw_output);
    BwImage edge_img(camera.edges);
    // using Sobel-operator to get gradient in x and y direction.
    for (int y=0; y<imageSize.height; y++)
    {
        //const uchar* edges_row = edges->data.ptr + y*edges->step;
        const short* dx_row = (const short*)(cam_left.dx->data.ptr + y*cam_left.dx->step);
        const short* dy_row = (const short*)(cam_left.dy->data.ptr + y*cam_left.dy->step);
        for (int x=0; x<imageSize.width; x++)
        {
            if (edge_img[y][x]!=0)
            {
                float vx, vy, sx, sy;
                vx = dx_row[x];
                vy = dy_row[x];
                float mag = sqrt(vx*vx+vy*vy);
                if (mag < 1)
                {
                    continue;
                }
                //assert( mag >= 1 );
		cout << "X: " <<  vx/mag << " Y: " << vy/mag << endl;
                sx = vx/mag;
                sy = vy/mag;
		output[y][x].r=(sx*128.0)+127;
		output[y][x].b=(sy*128.0)+127;
		//cvLine(raw_output, CvPoint(y,x),
		/*
                if (sx == 1)
                {
                    output[y][x].r=255;
                }
                else if ( sx == 0)
                {
                    output[y][x].r=255;
                    output[y][x].b=255;
                }
                else if ( sx == -1)
                {
                    output[y][x].b=255;
                }
                if (sy == 1)
                {
                    output[y][x].g=255;
                }
                else if (sy == 0)
                {
                    output[y][x].g=128;
                }
                else if (sy == -1)
                {
                    output[y][x].g=0;
                }
		*/
                //output[y][x].g=edges_row[x];
                //output[y][x].b=(sy*128 + 127);
                //cout << "sx: " << sx << " sy: " << sy << endl;
            }
        }
    }
    //cvReleaseMat(&dx);
    //cvReleaseMat(&dy);
    //cvReleaseImage(&edges);
}


void edge_block_matching(IplImage* raw_output)
{
    RgbImage output(raw_output);
    BwImage left_edges(cam_left.edges);
    BwImage right_edges(cam_right.edges);
    YuvImage left_input(cam_left.remapped.raw());
    YuvImage right_input(cam_right.remapped.raw());
    BwImage left_ch1(cam_left.ch1.raw());
    BwImage right_ch1(cam_right.ch1.raw());

    // rows are aligned
    for (int y=1; y<imageSize.height-1; y++)
    {
        const short* left_dx_row = (const short*)(cam_left.dx->data.ptr + y*cam_left.dx->step);
        const short* left_dy_row = (const short*)(cam_left.dy->data.ptr + y*cam_left.dy->step);
        const short* right_dx_row = (const short*)(cam_right.dx->data.ptr + y*cam_right.dx->step);
        const short* right_dy_row = (const short*)(cam_right.dy->data.ptr + y*cam_right.dy->step);
        for (int l_x=1; l_x<imageSize.width-1; l_x++)
        {
            if (left_edges[y][l_x]!=0)
            {
                int count=0;
                float left_vx, left_vy, left_sx, left_sy;
                left_vx = left_dx_row[l_x];
                left_vy = left_dy_row[l_x];
                float left_mag = sqrt(left_vx*left_vx+left_vy*left_vy);
                if (left_mag < 1)
                {
                    continue;
                }
                //assert( mag >= 1 );
                left_sx = cvRound(left_vx/left_mag);
                left_sy = cvRound(left_vy/left_mag);
                if (left_sx==-1)
                {
                    for (int r_x=l_x; r_x<imageSize.width-1; r_x++)
                    {
                        if (right_edges[y][r_x]!=0)
                        {
                            float right_vx, right_vy, right_sx, right_sy;
                            right_vx = right_dx_row[r_x];
                            right_vy = right_dy_row[r_x];
                            float right_mag = sqrt(right_vx*right_vx+right_vy*right_vy);
                            if (right_mag < 1)
                            {
                                continue;
                            }
                            //assert( mag >= 1 );
                            right_sx = cvRound(right_vx/right_mag);
                            right_sy = cvRound(right_vy/right_mag);
                            if (left_sy==right_sy && left_sx==right_sx)
                            {
                                if (block_match(left_ch1,right_ch1,y-left_sx,l_x+1,r_x+1,1,1)<=1)
                                {
                                    if (count<1)
                                    {
                                        setRGBfromYUV(left_input,output,y-left_sx,l_x+1);
                                        setRGBfromYUV(right_input,output,y-left_sx,r_x+1);
                                    }
                                    count++;
                                }
                            }
                        }
                    }
                }
            }

        }
    }

}

void edge_data_test(frame<RgbPixel> &camera, IplImage* raw_output)
{
    RgbImage output(raw_output);
    BwImage edge_img(camera.edges);
    YuvImage input(camera.remapped.raw());
    // using Sobel-operator to get gradient in x and y direction.
    for (int y=1; y<imageSize.height-1; y++)
    {
        //const uchar* edges_row = edges->data.ptr + y*edges->step;
        const short* dx_row = (const short*)(camera.dx->data.ptr + y*camera.dx->step);
        const short* dy_row = (const short*)(camera.dy->data.ptr + y*camera.dy->step);
        for (int x=1; x<imageSize.width-1; x++)
        {
            if (edge_img[y][x]!=0)
            {
                float vx, vy, sx, sy;
                vx = dx_row[x];
                vy = dy_row[x];
                float mag = sqrt(vx*vx+vy*vy);
                if (mag < 1)
                {
                    continue;
                }
                //assert( mag >= 1 );
                sx = cvRound(vx/mag);
                sy = cvRound(vy/mag);
                if (sx==0)
                {
                    // horizontal edge.
                    //cout << "Horizontal edge!" << endl;
                    setRGBfromYUV(input,output,y-1,x);
                    setRGBfromYUV(input,output,y+1,x);
                }
                if (sy==0)
                {
                    setRGBfromYUV(input,output,y,x-1);
                    setRGBfromYUV(input,output,y,x+1);
                }
                if (sx == -1 && sy == -1)
                {
                    setRGBfromYUV(input,output,y-1,x-1);
                    setRGBfromYUV(input,output,y+1,x+1);
                }
                if ((sx == -1 && sy == -1) || (sx == 1 && sy == 1))
                {
                    setRGBfromYUV(input,output,y-1,x-1);
                    setRGBfromYUV(input,output,y+1,x+1);
                }
                if ((sy == 1 && sx == -1) || (sy == -1 && sx == 1))
                {
                    setRGBfromYUV(input,output,y-1,x+1);
                    setRGBfromYUV(input,output,y+1,x-1);
                }
                /*
                                if (sx == 1)
                                {
                                    output[y][x].r=255;
                                }
                                else if ( sx == 0)
                                {
                                    output[y][x].r=255;
                                    output[y][x].b=255;
                                }
                                else if ( sx == -1)
                                {
                                    output[y][x].b=255;
                                }
                                if (sy == 1)
                                {
                                    output[y][x].g=255;
                                }
                                else if (sy == 0)
                                {
                                    output[y][x].g=128;
                                }
                                else if (sy == -1)
                                {
                                    output[y][x].g=0;
                                }
                  */
                //output[y][x].g=edges_row[x];
                //output[y][x].b=(sy*128 + 127);
                //cout << "sx: " << sx << " sy: " << sy << endl;
            }
        }
    }
    //cvReleaseMat(&dx);
    //cvReleaseMat(&dy);
    //cvReleaseImage(&edges);
}


void edge_matching_correspondence(IplImage *raw_left, IplImage* raw_output)
{
    RgbImage output(raw_output);
    CvMat *dx = 0, *dy = 0;
    BwImage edge_img(cam_left.edges);
    // using Sobel-operator to get gradient in x and y direction.
    //cvSobel( raw_input, left_dx, 1, 0, 3 );
    //cvSobel( raw_input, left_dy, 0, 1, 3 );
    for (int y=0; y<imageSize.height; y++)
    {
        //const uchar* edges_row = edges->data.ptr + y*edges->step;
        const short* dx_row = (const short*)(cam_left.dx->data.ptr + y*cam_left.dx->step);
        const short* dy_row = (const short*)(cam_left.dy->data.ptr + y*cam_left.dy->step);
        for (int x=0; x<imageSize.width; x++)
        {
            if (edge_img[y][x]!=0)
            {
                float vx, vy, sx, sy;
                vx = dx_row[x];
                vy = dy_row[x];
                float mag = sqrt(vx*vx+vy*vy);
                if (mag < 1)
                {
                    continue;
                }
                //assert( mag >= 1 );
                sx = cvRound(vx/mag);
                sy = cvRound(vy/mag);
                if (sx == 1)
                {
                    output[y][x].r=255;
                }
                else if ( sx == 0)
                {
                    output[y][x].r=255;
                    output[y][x].b=255;
                }
                else if ( sx == -1)
                {
                    output[y][x].b=255;
                }
                if (sy == 1)
                {
                    output[y][x].g=255;
                }
                else if (sy == 0)
                {
                    output[y][x].g=128;
                }
                else if (sy == -1)
                {
                    output[y][x].g=0;
                }
                //output[y][x].g=edges_row[x];
                //output[y][x].b=(sy*128 + 127);
                //cout << "sx: " << sx << " sy: " << sy << endl;
            }
        }
    }
    cvReleaseMat(&dx);
    cvReleaseMat(&dy);
    //cvReleaseImage(&edges);
}


void block_matching_correspondence(IplImage* raw_left, IplImage* raw_right, IplImage* raw_output)
{
    vector<unsigned int> disparities;
    disparities.assign(800,0);
    BwImage left_output_data(raw_left);
    BwImage right_output_data(raw_right);
    RgbImage output(raw_output);


    double lowest_mse=100000000.0; // using huge number.
    double mse;
    int best_xl=0;
    for (int y=0; y<imageSize.height-10; y+=10)
    {
        for (int x=0; x<imageSize.width-10; x+=10)
        {
            if (right_output_data[y][x]==0)
            {
                continue;
            }
            lowest_mse=100000000000.0;
            for (int i=0; i<imageSize.width-10; i++)
            {
                mse=block_match(left_output_data, right_output_data, y, x, x+i, 10, 10);
                if (mse<lowest_mse)
                {
                    best_xl=x+i;
                    lowest_mse=mse;
                }
            }
            // drawing correspondence for block.
            int disp=abs(x-best_xl);
            disparities[disp]++;
            int distance=(int) disparity_to_distance(disp);
            //cout << y << " " << x << " MSE: " << lowest_mse << " disp: " << disp << endl;
            //block_draw(mse_output,y,x,5,5,lowest_mse*100);
            if (lowest_mse<10.0)
            {
                //cout << lowest_mse << endl;
                block_draw(output,y,x,10,10,distance);
            } else {
                block_draw(output,y,x,10,10,-1);
            }
        }
        //cvShowImage("disparity",raw_output);
    }
    int max=0,max_disp=0;
    for (int i=1; i<disparities.size();i++)
    {
        if (max<disparities[i])
        {
            max=disparities[i];
            max_disp=i;
        }
    }
    /*
    for (int i=0; i<disparities.size(); i+=5)
    {
        int values=0;
        for (int j=0; j<5; j++)
        {
            values+=disparities[i+j];
        }
        block_draw(output,460,i,20,5,values);
    }
    */
    double distance=disparity_to_distance(max_disp);
    double distance_over=disparity_to_distance(max_disp+1)-distance;
    double distance_under=disparity_to_distance(max_disp-1)-distance;
    cout << "disparity: " << max_disp << "(values: " << max << " )" << endl;
    cout << "Z:" << disparity_to_distance(max_disp) << endl;
/*
    for (int y=0; y<imageSize.height; y++)
    {
        //int color=255;
        // epipolar lines
        int last_xl=0;
        for (int x_r=0; x_r<imageSize.width; x_r++)
        {
            if (right_output_data[y][x_r] != 0)
            {
                for (int x_l=last_xl+1; x_l<imageSize.width; x_l++)
                {
                    if (left_output_data[y][x_l] != 0)
                    {
                        if (abs(left_output_data[y][x_l]-right_output_data[y][x_r])<20.0)
                        {
                            int disp=abs(x_r-x_l);
                            //output[y][x_l].r=CLIP(disp,0,255);
                            output[y][x_r].r=CLIP(disp,0,255);
                            last_xl=x_r;
                            x_r++;
                        }
                    }
                }
            }
        }
    }
    */
}

// this routine takes two run length encoded segmentation images, and tries to find correspondences.

void RLE_match(vector< vector< run_len > > &left_rle, vector< vector< run_len > > &right_rle, RgbImage& output, int size_threshold, int color_threshold)
{
    RgbPixel green;
    RgbPixel black;
    black.r=0;
    black.g=0;
    black.b=0;
    green.r=100;
    green.g=255;
    green.b=50;
    for(int y=0; y<left_rle.size(); y++) {
        //printf("%d: ",y);
        for(int r_x=1; r_x<right_rle[y].size(); r_x++) {
            run_len right = right_rle.at(y).at(r_x);
            if(color_difference(right.colorval,green)<100 || color_difference(right.colorval,black)<100) { continue; }
            for(int l_x=1; l_x<left_rle[y].size(); l_x++) {
                run_len left = left_rle.at(y).at(l_x);
                if(left.pos<right.pos) { continue; }
                //printf("%d ",color_difference(left.colorval, right.colorval));
                // compare size (sz) and colorval.
                if(abs(left.sz-right.sz)<size_threshold && color_difference(left.colorval, right.colorval)<color_threshold ) {
                    for(int i=0; i<(left.sz+right.sz)/2; i++) {
                        //output[y][i+left.pos]=right.colorval;
                        int d = abs(left.pos-right.pos);
                        int pos = (left.pos+right.pos)/2;
                        if(d>=0 && d<200) {
                            output[y][i+pos].b=min(abs(left.sz-right.sz),255);
                            output[y][i+pos].r=min((int) color_difference(left.colorval, right.colorval),255);
                            output[y][i+pos].g=min(d,255);
                        }
                    }
                }
            }
        }
        //printf(" \n");
    }

}


void thresholding_test(IplImage* raw_input, IplImage* histogram)
{
    calculate_histogram(raw_input, histogram);
    //cvShowImage("left",raw_input);
    //cvShowImage("right",right_input);
}


// this simply combines the left and right image into an anaglyphic stereo image.
// supports B&W and RGB images.
void anaglyphic_stereo(IplImage* raw_output)
{
        RgbImage left_data(cam_left.remapped);
        RgbImage right_data(cam_right.remapped);
        RgbImage output(raw_output);
        for (int y=0; y<imageSize.height; y++)
        {
            for (int x=0; x<imageSize.width; x++)
            {
                output[y][x].r=CLIP(right_data[y][x].r*1.1,0,255);
                output[y][x].g=CLIP(left_data[y][x].g*1.1,0,255);
                output[y][x].b=CLIP(left_data[y][x].b*1.1,0,255);
            }
        }
}



