#ifndef GROUND_TRUTH_HPP
#define GROUND_TRUTH_HPP
#include <iostream>
#include <cstdio>
#include <cmath>
#include "segmentation.hpp"

typedef struct {
	double horizontal;
	double vertical;
	double diagonal;
	unsigned int res_vertical;
	unsigned int res_horizontal;
} sensor_size;

#ifdef STANDALONE
typedef struct {
    int y;
    int x;
} coord;
#endif

double angle_from_position_and_orientation(int y, int x, double orientation);
CvPoint position_from_distance_and_orientation(double orientation, double y, double x);

class ground_truth {
	private:
		double downwards_angle;
		double view_height;
		double focal_length;
		sensor_size ccd;
		double corn_height;
		double vertical_resolution;
		double horizontal_resolution;
		double field_of_view_CCD(double sensor_size, double focal_length);
		double pixel_size_at_distance(double distance);
    public:
        ground_truth();
        double getFL();
        void setFL(double value);
        void setViewHeight(double meters);
        double getViewHeight();
        void setDownwardsAngle(double value);
        double getDownwardsAngle();
        double getFOV_vertical();
        double getFOV_horizontal();
        double getFOV_diagonal();
        double distance_from_vertical_index(int vertical_index);
        double distance_from_vertical_angle(double angle);
        double distance_from_vertical_angle(double angle, double height);
        int horizontal_index_from_angle(double);
        double vertical_angle_from_distance(double distance, double height);
        double vertical_angle_from_distance(double distance);
        CvPoint image_position_from_real(double distance, double angle, int height);
        CvPoint image_position_from_real_coord(int y, int x, double orientation, int z);
        double error_from_vertical_index(int vertical_index);
        int ccd_coverage_in_pixels(double distance, double size);
        int ball_size_from_distance(double distance);
        coord position_from_ground_pixel(int y, int x);
        double horizontal_position_from_distance(double distance, int x);
};
#endif // GROUND_TRUTH_HPP
