#include "individual_processing.hpp"
#include "recognition/game_elements.hpp"
#include "board.hpp"
#include "state.hpp"
#include "communication.hpp"
using namespace std;

// global variables defined here to avoid multiple definitions.

// global mutexes, barriers etc.
pthread_mutex_t display = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t enumeration_mutex = PTHREAD_MUTEX_INITIALIZER;
//volatile int cameras_ready;
int rc;

// barriers
pthread_barrier_t barrier_before_processing, barrier_after_processing, barrier_correspondence;

frame<RgbPixel> cam_left;
frame<RgbPixel> cam_right;

void init_matrices(frame<RgbPixel> &camera, int channels)
{
    camera.dx = cvCreateMat( imageSize.height,imageSize.width, CV_16SC1 );
    camera.dy = cvCreateMat( imageSize.height,imageSize.width, CV_16SC1 );
    //camera.storage = cvCreateMemStorage(0);
    //camera.circles = cvCreateSeq( CV_32FC3, sizeof(CvSeq),sizeof(float)*3, (CvMemStorage*) cam_left.storage );
    //camera.contours = cvCreateSeq(CV_SEQ_ELTYPE_POINT, sizeof(CvSeq), sizeof(CvPoint) , (CvMemStorage*) camera.storage);
    camera.input = cvCreateImage( imageSize, 8, channels);
    camera.intensity = cvCreateImage( imageSize, 8, 1);
    camera.remapped = cvCreateImage( imageSize, 8, channels);
    camera.temp = cvCreateImage( imageSize, 8, 1);
    camera.output = cvCreateImage( imageSize, 8, channels);
    camera.ch1 = cvCreateImage( imageSize, 8, 1);
    camera.ch2 = cvCreateImage( imageSize, 8, 1);
    camera.ch3 = cvCreateImage( imageSize, 8, 1);
    camera.diff = cvCreateImage( imageSize, 8, channels);
    camera.edges = cvCreateImage( imageSize, 8, 1);
    camera.timestamp = 0;
    camera.ready=0;
    //camera.mutex = PTHREAD_MUTEX_INITIALIZER;
}

void release_matrices(frame<RgbPixel> &camera)
{
    cvReleaseMat(&camera.dx);
    cvReleaseMat(&camera.dy);
    //cvReleaseStorage(&camera.storage);
    //cvReleaseSequence(&camera.circles);
    /*
    cvReleaseImage(&camera.input);
    cvReleaseImage(&camera.remapped);
    cvReleaseImage(&camera.temp);
    cvReleaseImage(&camera.output);
    cvReleaseImage(&camera.ch1);
    cvReleaseImage(&camera.ch2);
    cvReleaseImage(&camera.ch3);
    cvReleaseImage(&camera.diff);
    cvReleaseImage(&camera.edges);
    */
    camera.ready=0;
}

void merge_images(IplImage *raw_input, IplImage *raw_output)
{
    BwImage input(raw_input);
    BwImage output(raw_output);
    for (int y=0; y<imageSize.height; y++)
    {
        for (int x=0; x<imageSize.width; x++)
        {
            if (input[y][x]!=0)
            {
                output[y][x]=input[y][x];
            }
        }
    }
}

void detect_red(RgbPixel& input, RgbPixel& output)
{
    double ratio_rg = input.r/(double) input.g;
    double ratio_rb = input.r/(double) input.b;
    double ratio_gb = input.g/(double) input.b;
    if (ratio_rg >= 1.3 && ratio_rb >= 1.3)
    {
        //output.r=255;
        //output.g=0;
        //output.b=0;
        output=input;
    }  /*else if (ratio_rg < 0.75) {
        output.r=0;
        output.g=255;
        output.b=0;
    } else if (ratio_rb >= 2.5 && ratio_gb >= 2.5){
        output.r=255;
        output.g=255;
        output.b=0;
    } else if (intensity(input)>= 100.0) {
        output.r=255;
        output.g=255;
        output.b=255;
    } */
}


// rectification, color conversion and thresholding is done here.
// anything that can be done on individual images should preferably be done here.
void compute_on_frame(frame<RgbPixel> &camera, double incoming_time, double conversion_time)
{
    double remap_start, remap_end;
    double computation_start, computation_end;
    //IplImage *last_frame = (IplImage*) cvClone(camera.remapped);
    remap_start = get_time();
    rc = pthread_barrier_wait(&barrier_correspondence);

    if (options.option!=0 && options.option != 15 && options.option!=13)
    {
        wrapper::rectify_image(camera.input, camera.remapped, camera.mx, camera.my );
    }
    else
    {
        wrapper::copy_image(camera.input, camera.remapped);
    }

    remap_end = get_time();
    //cvSub(last_frame, camera.remapped, camera.diff);
    //cvReleaseImage(&last_frame);
    // just for testing.
    //RGB_threshold_on_HSV(camera.input,camera.remapped, imageSize);

    computation_start = get_time();
    if (options.option==4)
    {
        wrapper::gaussian_blur(camera.remapped, camera.remapped, 9);
        if (camera.circles) cvClearSeq(camera.circles);
        HoughOptimized(camera.remapped.raw(),2.0,20,20,100,200,40,camera.circles,10);
    }
    if (options.option==5)
    {
        //split_3channels(camera.remapped, camera.ch1, camera.ch2, camera.ch3);
        /*
        for(int y=0; y<imageSize.height; y++) {
                for(int x=0; x<imageSize.width; x++) {
                    normalize(camera.remapped[y][x],100.0);
                }
            } */

        //detect_colors(camera.remapped, camera.output);
        wrapper::convert_rgb2gray(camera.input, camera.ch1);
        cvSmooth(camera.ch1.raw(), camera.ch1.raw(), CV_GAUSSIAN, 9, 9);
        if (camera.circles) cvClearSeq(camera.circles);
        HoughOptimized(camera.ch1.raw(),2.0,20,10,60,120,60,camera.circles,50); // min distance is approx. 28 on small table, 150+ up close.
    }
    if (options.option==2 || options.option==9 || options.option==10 || options.option==13)
    {
        // setting ROI to avoid edges from remapping.
        if (options.color==0)
        {
            // does not support BW.
            //wrapper::copy_image(camera.remapped,camera.ch1);
        }
        else
        {
            split_3channels(camera.remapped, camera.ch1, camera.ch2, camera.ch3);
        }
        wrapper::gaussian_blur(camera.ch1,camera.ch1,3);
        wrapper::gradient_horizontal( camera.ch1, camera.dx );
        wrapper::gradient_vertical( camera.ch1, camera.dy );
        wrapper::set_ROI(camera.ch1.raw(), 30, 25, imageSize.width, imageSize.height);
        wrapper::set_ROI(camera.edges.raw(), 30, 25, imageSize.width, imageSize.height);
        wrapper::find_edges(camera.ch1, camera.edges);
        wrapper::reset_ROI(camera.ch1.raw());
        wrapper::reset_ROI(camera.edges.raw());
    }
    if (options.option==17)
    {
        //wrapper::gaussian_blur(camera.remapped, camera.remapped, 5);
        RGB_Gradient(camera.remapped.raw(), camera.dx, 2);
        /*
                    for(int y=0; y<imageSize.height; y++) {
                        for(int x=0; x<imageSize.width; x++) {
                            normalize(camera.remapped[y][x],100.0);
                        }
                    }
        */
        //detect_colors(camera.remapped, camera.remapped);
        color_encode(camera.remapped, camera.dx, camera.output,40, 40);
        for (int i=0; i<2; i++)
        {
            fill_holes(camera.output,camera.diff,10);
            smooth_region(camera.output);
        }
        //region_fix(camera.output,camera.remapped, camera.diff, 100);
    }
    if (options.option==18)
    {
        //wrapper::gaussian_blur(camera.remapped, camera.remapped, 3);
        cvSmooth(camera.remapped.raw(),camera.remapped.raw(),CV_MEDIAN,3,3);
        RGB_Gradient(camera.remapped.raw(), camera.dx, 2);
        /*
                   for(int y=0; y<imageSize.height; y++) {
                        for(int x=0; x<imageSize.width; x++) {
                            normalize(camera.remapped[y][x],100.0);
                        }
                    }
        */
        color_encode(camera.remapped, camera.dx, camera.output,40, 40);
        for(int i=0; i<2; i++) {
            //fill_holes(camera.output,camera.diff,40);
            smooth_region(camera.output);
        }
        //region_fix(camera.output, camera.remapped, camera.diff, 100);
    }

    if (options.option==19)
    {
        //color_encode(camera.remapped, camera.dx, camera.output,70, 70);
        RGB_Gradient(camera.remapped.raw(), camera.dx,2);
        MatShort gradient_rgb(camera.dx);
        for (int y=0; y<imageSize.height; y++)
        {
            for (int x=0; x<imageSize.width; x++)
            {
                //if(gradient_rgb[y][x] < 60) {
                //printf("gradient: %d\n", gradient_rgb[y][x]);
                normalize_fix(camera.remapped[y][x],100.0);
                /*
                } else {
                camera.remapped[y][x].r=0;
                camera.remapped[y][x].g=0;
                camera.remapped[y][x].b=255;
                }
                */
            }
        }
        /*
        for(int y=0; y<imageSize.height; y++) {
            for(int x=0; x<imageSize.width; x++) {
                    detect_red(camera.remapped[y][x],camera.output[y][x]);
            }
        }
        */
        //wrapper::convert_rgb2gray(camera.remapped,camera.temp);
        //cvLaplace(camera.temp.raw(),camera.dx,3);
        //wrapper::convert_rgb2gray(camera.output,camera.temp);
        //wrapper::gradient(camera.temp,camera.dx);
        //wrapper::find_edges(camera.temp,camera.edges);
        //*/
        //border_detect(camera.output,camera.remapped);
        //RGB_gradient_ignore_black(camera.output, camera.dx);
        /*
        for(int y=0; y<imageSize.height; y++)  {
            for(int x=0; x<imageSize.width; x++) {
                if(camera.diff[y][x].r!=0) {
                    camera.output[y][x]=camera.diff[y][x];
                }
            }
        }
        */
        //wrapper::copy_image(camera.diff,camera.output);

        //cvSmooth(camera.output.raw(),camera.output.raw(),CV_MEDIAN,3,3);
        /*

        */
    }
    if (options.option == 21)
    {
        find_game_elements_individual(camera);
    }
    computation_end = get_time();
    rc = pthread_barrier_wait(&barrier_after_processing);
    /*
    if (*frame_count % 10 == 0)
    {
        pthread_mutex_lock(&display);
        cout << ": " << (*frame_count)/(incoming_time-zero) << "fps Frame: " << *frame_count << " (" << incoming_time-zero << " s)"
        << " Conversion: " << conversion_time*1000.0
        << "ms Remap: " << (remap_end-remap_start)*1000.0
        << "ms Computation: " << (computation_end-computation_start)*1000.0 << " ms"<< endl;
        pthread_mutex_unlock(&display);
    }
    */
}


// this function is called for every incoming frame.
void new_frame( int camera_id, unsigned char *input_buffer)
{
    double conversion_start, conversion_end;
    double incoming_time = get_time();
    //long *frame_count = *frame_count + 1;
    int state = getGlobalState();
    if (state == STATE_RUNNING)
    {
        rc = pthread_barrier_wait(&barrier_before_processing);
        conversion_start = get_time();
        if (camera_id == LEFT_CAMERA)
        {
            color_convert(input_buffer,cam_left.input.raw(), cam_left.intensity,imageSize);
            conversion_end = get_time();
            compute_on_frame(cam_left,incoming_time,(conversion_end-conversion_start));
        }
        else
        {
            color_convert(input_buffer,cam_right.input.raw(), cam_right.intensity, imageSize);
            conversion_end = get_time();
            compute_on_frame(cam_right,incoming_time,(conversion_end-conversion_start));
        }
    }
    if (state == STATE_CYLINDER_CONFIG)
    {
        if (our_color == 1 && camera_id == RIGHT_CAMERA)
        {
            color_convert(input_buffer,cam_right.input.raw(), cam_right.intensity,imageSize);
            conversion_end = get_time();
            wrapper::rectify_image(cam_right.input, cam_right.remapped, cam_right.mx, cam_right.my );
            setGlobalState(STATE_RUNNING);
            hard_coded_cylinders(cam_right.remapped,cam_right.temp);
            wrapper::save_image(cam_right.remapped.raw(),"start_config_yellow_rectified.png");
        }
        if (our_color == 0 && camera_id == LEFT_CAMERA)
        {
            color_convert(input_buffer,cam_left.input.raw(), cam_left.intensity,imageSize);
            conversion_end = get_time();
            wrapper::rectify_image(cam_left.input, cam_left.remapped, cam_left.mx, cam_left.my );
            setGlobalState(STATE_RUNNING);
            hard_coded_cylinders(cam_left.remapped,cam_left.temp);
            wrapper::save_image(cam_left.remapped.raw(),"start_config_blue_rectified.png");
        }
    }
}
