#ifndef WRAPPER_HPP
#define WRAPPER_HPP
#include "image.hpp"
#include "canny.hpp"

namespace wrapper {

// input: 8-bit B&W Image
// output: 16-bit Matrix
void gradient_horizontal(BwImage& input, CvMat* raw_output)
{
    // parameters: x-order, y-order, aperture size.
    cvSobel( input.raw(), raw_output, 1, 0, 3 );
}

// input: 8-bit B&W Image
// output: 16-bit Matrix
void gradient_vertical(BwImage& input, CvMat* raw_output)
{
    // parameters: x-order, y-order, aperture size.
    cvSobel( input.raw(), raw_output, 0, 1, 3 );
}

void gradient(BwImage& input, CvMat* raw_output)
{
    cvSobel( input.raw(), raw_output, 1, 1, 3 );
}

// input == output, mask size.
template<class T>
void gaussian_blur(Image<T>& input, Image<T>& output, unsigned char size)
{
    cvSmooth(input.raw(), output.raw(), CV_GAUSSIAN, size, size);
}

template<class T>
void rectify_image(Image<T>& input, Image<T>& output, CvMat* horizontal_correction, CvMat* vertical_correction)
{
    cvRemap(input.raw(), output.raw(), horizontal_correction, vertical_correction );
}

template<class T>
void copy_image(Image<T>& input, Image<T>& output)
{
    cvCopy(input.raw(),output.raw(),0);
}

void zero_fill(IplImage* raw_input)
{
    cvSet(raw_input,cvScalar(0,0,0));
}

void save_image(IplImage* raw_input, std::string filename)
{
    cvSaveImage(filename.c_str(), raw_input);
}

void set_ROI(IplImage* raw_input, int x1, int y1, int x2, int y2)
{
    cvSetImageROI(raw_input,cvRect(x1, y1, x2, y2));
}

void reset_ROI(IplImage* raw_input)
{
    cvResetImageROI(raw_input);
}

void convert_rgb2gray(RgbImage& input, BwImage& output)
{
    cvCvtColor(input.raw(),output.raw(),CV_BGR2GRAY);
}

void scale_8bit(CvMat* raw_input, BwImage& output) {
    cvNormalize(raw_input,output.raw(),0,255,CV_MINMAX);
}

void scale_8bit(BwImage& input, BwImage& output) {
    cvNormalize(input.raw(), output.raw(), 0, 255, CV_MINMAX);
}

template<class T>
void show_image(std::string window, Image<T>& img) {
    cvShowImage(window.c_str(), img.raw());
}

// input: 8-bit, output: 8-bit.
void find_edges(BwImage& input, BwImage& output)
{
    // parameters: threshold for edge, threshold for followed edges, aperture size for Sobel operator
    cvCanny(input.raw(),output.raw(), 20, 40, 3); // 60, 80, 3
}

void find_edges(BwImage& input, BwImage& output, CvMat* dx, CvMat* dy)
{
    Canny detector(800,600);
    detector.set_threshold(40,80);
    detector.perform(input.raw(),dx,dy);
    copy_image(input, output);
}

}
#endif
