#ifndef __CAN_MESSAGES_H__
#define __CAN_MESSAGES_H__

/**
 *  CAN Base addresses
 *  (Each module has its own base address and the
 *   modules are assigned an address range of 1byte)
 */

#define CAN_BASE_ADDR_SYS                   0x000
#define CAN_BASE_ADDR_SENSOR				0x100
#define CAN_BASE_ADDR_MOTORCONTROL			0x200
#define CAN_BASE_ADDR_EIT                   0x300

/**
 * CAN SYS Messages
 * (See the documentation for the switch panel)
 */

//#define CAN_SYS_MSG_GAME_STOP    			0x05 + CAN_BASE_ADDR_SYS
//#define CAN_SYS_MSG_LOW_BATTERY			0x06 + CAN_BASE_ADDR_SYS

#define CAN_SYS_MSG_START_WIRE	 		    0x10 + CAN_BASE_ADDR_SYS
#define CAN_SYS_MSG_RESET_BUTTON 		    0x11 + CAN_BASE_ADDR_SYS
#define CAN_SYS_MSG_BLUE_TEAM_BUTTON	    0x15 + CAN_BASE_ADDR_SYS
#define CAN_SYS_MSG_YELLOW_TEAM_BUTTON	    0x16 + CAN_BASE_ADDR_SYS

#define CAN_SYS_MSG_RESET_LED 		    	0x21 + CAN_BASE_ADDR_SYS
#define CAN_SYS_MSG_BLUE_TEAM_LED	    	0x25 + CAN_BASE_ADDR_SYS
#define CAN_SYS_MSG_YELLOW_TEAM_LED	    	0x26 + CAN_BASE_ADDR_SYS



/**
 * 	CAN MOTOR CONTROL Messages
 *	(see the documentation for functionality)
 */

#define CAN_MOTOR_MSG_PID                   0x00 + CAN_BASE_ADDR_MOTORCONTROL
#define CAN_MOTOR_MSG_SPEED_REFERENCE       0x01 + CAN_BASE_ADDR_MOTORCONTROL
//#define CAN_MOTOR_MSG_CONFIG              0x02 + CAN_BASE_ADDR_MOTORCONTROL


#define CAN_MOTOR_MSG_STATUS                0x0A + CAN_BASE_ADDR_MOTORCONTROL
#define CAN_MOTOR_MSG_ERROR                 0x0B + CAN_BASE_ADDR_MOTORCONTROL
#define CAN_MOTOR_MSG_SPEED_MEASUREMENT     0x0C + CAN_BASE_ADDR_MOTORCONTROL
#define CAN_MOTOR_MSG_CURRENT_MEASUREMENT   0x0D + CAN_BASE_ADDR_MOTORCONTROL


/**
 * 	CAN SENSOR Messages
 *	(see the documentation for functionality)
 */
/*CONFIG register*/
#define CAN_SENSOR_MSG_CONFIG			    0x00 + CAN_BASE_ADDR_SENSOR

/*ODOMETRY*/
#define CAN_SENSOR_MSG_ODOMETRY 		    0x10 + CAN_BASE_ADDR_SENSOR
#define CAN_SENSOR_MSG_ODOMETRY_SPEED	    0x11 + CAN_BASE_ADDR_SENSOR
#define CAN_SENSOR_MSG_ODOMETRY_RESET	    0x12 + CAN_BASE_ADDR_SENSOR


/*INERTIAL*/
#define CAN_SENSOR_MSG_ACCELEROMETER	    0x21 + CAN_BASE_ADDR_SENSOR
#define CAN_SENSOR_MSG_RATE_GYRO		    0x22 + CAN_BASE_ADDR_SENSOR
#define CAN_SENSOR_MSG_TEMPERATURE 		    0x23 + CAN_BASE_ADDR_SENSOR
#define CAN_SENSOR_MSG_VOLTAGE 			    0x24 + CAN_BASE_ADDR_SENSOR


/*ANALOG AND DIGITAL I/O*/
#define CAN_SENSOR_MSG_ANALOG			    0x31 + CAN_BASE_ADDR_SENSOR
#define CAN_SENSOR_MSG_DIGITAL_IN		    0x32 + CAN_BASE_ADDR_SENSOR
#define CAN_SENSOR_MSG_DIGITAL_OUT		    0x33 + CAN_BASE_ADDR_SENSOR
#define CAN_SENSOR_MSG_DIP_SWITCH		    0x34 + CAN_BASE_ADDR_SENSOR


/*WALLE*/
#define CAN_SENSOR_MSG_WALLE_ENEMY_POS		0x41 + CAN_BASE_ADDR_SENSOR
#define CAN_SENSOR_MSG_WALLE_TOWER_POS		0x42 + CAN_BASE_ADDR_SENSOR
#define CAN_SENSOR_MSG_WALLE_DISPENSER_POS	0x43 + CAN_BASE_ADDR_SENSOR
#define CAN_SENSOR_MSG_WALLE_ERROR		    0x44 + CAN_BASE_ADDR_SENSOR


#define CAN_SENSOR_MSG_WALLE_LOCK_EMEMY		0x4A + CAN_BASE_ADDR_SENSOR
#define CAN_SENSOR_MSG_WALLE_FIND_TOWERS	0x4B + CAN_BASE_ADDR_SENSOR
#define CAN_SENSOR_MSG_WALLE_FIND_DISPENSER 0x4C + CAN_BASE_ADDR_SENSOR
#define CAN_SENSOR_MSG_WALLE_RESET_ANGLE  	0x4D + CAN_BASE_ADDR_SENSOR
#define CAN_SENSOR_MSG_WALLE_STOP		    0x4E + CAN_BASE_ADDR_SENSOR


/*STATUS*/
#define CAN_SENSOR_MSG_STATUS			    0x50 + CAN_BASE_ADDR_SENSOR


/**
 * CAN EIT Messages
 * (See the documentation for EIT module)
 */

/*To EiT module*/
#define CAN_EIT_MSG_EIT_STOP				0x01 + CAN_BASE_ADDR_EIT
#define CAN_EIT_MSG_START					0x02 + CAN_BASE_ADDR_EIT
#define CAN_EIT_MSG_DELIVER					0x03 + CAN_BASE_ADDR_EIT
#define CAN_EIT_MSG_PICK_TOMATO				0x04 + CAN_BASE_ADDR_EIT
#define CAN_EIT_MSG_PREPARE_FOR_ORANGE		0x05 + CAN_BASE_ADDR_EIT
#define CAN_EIT_MSG_PICK_ORANGE				0x06 + CAN_BASE_ADDR_EIT

/*From EiT module*/
#define CAN_EIT_MSG_TOMATO_PICKED			0x11 + CAN_BASE_ADDR_EIT
#define CAN_EIT_MSG_CORN_PICKED				0x12 + CAN_BASE_ADDR_EIT
#define CAN_EIT_MSG_ORANGE_PICKED			0x13 + CAN_BASE_ADDR_EIT
#define CAN_EIT_MSG_BLOCKING_CORN			0x14 + CAN_BASE_ADDR_EIT
#define CAN_EIT_MSG_DELIVERY_DONE			0x15 + CAN_BASE_ADDR_EIT
#define CAN_EIT_MSG_INITIALIZED                         0x16 + CAN_BASE_ADDR_EIT

#endif
