#ifndef __COMMUNICATION_STRUCTS_H__
#define __COMMUNICATION_STRUCTS_H__

#include <sys/time.h>
#include <libpcan.h>

/**
 * GATEWAY(Sensor) to NAVSYS
 */

typedef struct{
    int left;
    int right;
    struct timeval timestamp;
}odometry_t;

typedef struct{
    int x;
    int y;
    int z;
    double temperature;
    double voltage;
    struct timeval timestamp;
}inertial_t;

typedef struct{
    unsigned short channel[4];
}sensor_analog_t;

typedef struct{
    unsigned char pin[4];
}sensor_digital_in_t;

typedef struct{
    /*Enable sensor flag (using bit fields)*/
    unsigned char odometry          :1;
    unsigned char accelerometer     :1;
    unsigned char gyro              :1;
    unsigned char analog            :1;
    unsigned char digital           :1;
    unsigned char odometry_speed    :1;
    unsigned char                   :0;
    unsigned char odometry_mode;    //1:1x, 2:2x 4:4x mode
}sensor_config_t;

/**
 * NAVSYS to GATEWAY(Motor Control)
 */
typedef struct{
    unsigned char P;
    unsigned char I;
    unsigned char D;
    unsigned char maxChange;
}motor_pid_t;

typedef struct{
    char motor1;
    char motor2;
}motor_speed_ref_t;


typedef struct{
    unsigned char pin[8];
}sensor_digital_out_t;

/**
 * GATEWAY(Motor Control) to NAVSYS
 */
typedef unsigned char motor_status_t;
typedef unsigned char motor_error_t;

typedef struct{
    char motor1;
    char motor2;
}motor_speed_t;

typedef struct{
    unsigned char motor1;
    unsigned char motor2;
}motor_current_t;

/**
 * AI to CVSYS
 */
typedef struct {
	short y;
	short x;
}coord;

typedef struct {
	char objType; //The type of object that was picked: c is corn, o is orange, t is tomato.
	unsigned char objIndex;
}objectPickup;


/**
 * CVSYS to AI
 */

typedef struct {
	short y;
	short x;
	char color; // 0 - hvit, 1 - svart, -1 udefinert
}cylinder_struct;


typedef struct {
	cylinder_struct cylinder[18];
	unsigned char counter; //Counts how many pickup messages CvSys has received. Used for synchronization.
}cylinder_config;

typedef struct {
	coord tomato[14]; // -1, -1
	coord orange[12]; // -1, -1
	unsigned char counter; //Counts how many pickup messages CvSys has received. Used for synchronization.
}ball_positions;


typedef struct {
	unsigned char camera_status; // 0 - FEIL, 1 - DELVIS OK, 2 - ALT OK
}cvsys_systemcheck;

/**
 * AI to navsys
 */

typedef struct{
    short x;
    short y;
    double psi;
    }starting_position;

typedef struct {
	short x; // i mm
	short y;
	double theta; // i radianer 0-2*PI (0 er lang x-aksen, pi/2 langs y-aksen)
	bool should_correct_angle; // true hvis vinkel er viktig.
	bool top_of_hill;
	int circle; // hvis roboten skal rundt en hindring. sendes p� f�rste waypoint.
	bool stop_on_wp; // ved levering, eventuelt tomatplukking
	bool reverse; //Robot drives in reverse to position.
}waypoint_struct;

typedef struct {
	waypoint_struct wp_list[2];
}waypoint_list;


/**
* navsys to AI
*/

typedef struct {
	short x;
	short y;
	double psi;
	timeval timestamp;
	double speed;
}robot_position;

typedef struct {
	short x;
	short y;
}opponent_position;


/**
 *	AI to GATEWAY(EiT)
 */

typedef struct{
    /*Enable sensor flag (using bit fields)*/
    unsigned char tomatoes          :1;
    unsigned char oranges	        :1;
    unsigned char corns             :1;
}eit_delivery_t;


#endif
