#ifndef IMAGE_HPP
#define IMAGE_HPP
#include <cv.h>
#include <highgui.h>
//#include <cxmisc.h>
#include <cxcore.h>
// includes for various
#include <vector>
#define _MULTI_THREADED
#include <pthread.h>
#include <string>


// template for easier pixel element access.
template<class T> class Image
{
	private:
        IplImage* imgp;
	public:
        Image(IplImage* img=0) {imgp=img;}
        ~Image(){imgp=0;}
        void operator=(IplImage* img) { imgp=img; }
        inline T* operator[](const int rowIndx) {
            return ((T *)(imgp->imageData + rowIndx*imgp->widthStep));
        }
        int getWidth() {
            return imgp->width;
        }
        void create(int width, int height, int bits, int channels) {
            imgp= cvCreateImage(cvSize(height,width), bits, channels);
        }
        int getHeight() {
            return imgp->height;
        }
        CvSize getSize() {
            return cvSize(imgp->width,imgp->height);
        }
        IplImage* raw() {
            return imgp;
        }
        void save(std::string filename);
        void zero_fill();
};


// template for easier matrix element access.
template<class T> class Matrix
{
    private:
        CvMat* matp;
    public:
        Matrix(CvMat* mat=0) { matp=mat; }
        ~Matrix(){ matp=0; }
        void operator=(CvMat* mat) { matp=mat; }
        inline T* operator[](const int rowIndx) {
            return ((T *)(matp->data.ptr + rowIndx*matp->step));
        }
        CvMat* raw() {
            return matp;
        }
        int getHeight() {
            return matp->height;
        }
        int getWidth() {
            return matp->width;
        }
};


typedef struct{
	unsigned char b,g,r;
} RgbPixel;

// u = chromatic blue, so mapped to blue.
// y = intensity, mapped to green where it is most visible.
// v = chromatic red, so mapped to red.
typedef struct{
	unsigned char u,y,v;
} YuvPixel;


typedef struct {
    int r,g,b;
} RgbPixelInt;

typedef struct{
	float b,g,r;
} RgbPixelFloat;

typedef struct{
    unsigned char b,g,r,y;
} RgbYPixel;

typedef Image<RgbPixel>       RgbImage;
typedef Image<YuvPixel>       YuvImage;
typedef Image<RgbPixelFloat>  RgbImageFloat;
typedef Image<RgbYPixel>      RgbYImage;
typedef Image<unsigned char>  BwImage;
typedef Image<unsigned short> BwImageShort;
typedef Image<float>          BwImageFloat;
typedef Matrix<short>         MatShort;

// options
typedef struct {
	int color, option, sync_error, input;
} opt_struct;

template<class T>
class frame {
    public:
        Image<T> input, output, remapped, diff;
        BwImage edges, temp, ch1, ch2, ch3, intensity;
        CvSeq *circles;
        CvMat *dx, *dy;
        CvMemStorage *storage;
        pthread_mutex_t mutex;
        CvMat *mx, *my;
        bool ready;
        int time;
        timeval initial_time;
};


#endif // IMAGE_HPP
