extern "C"
{
#include "urg_ctrl.h"
#include "delay.h"
}
#include <iostream>
#include "image.hpp"
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "wrapper.hpp"

using namespace std;


// communication
#include "../gateway/messageQueue.h"
#include "../gateway/communication_structs.h"
//#include "../../../gateway/communication.h"
#include "../gateway/posix_messages.h"
#include <ctime>


#include <string>
#include <sstream>
#include <cstdlib>
#include <cstring>
using namespace std;

// Global data structures
MessageQueue cvsysQueue(PMQ_CVSYS);
MessageQueue aiQueue(PMQ_AI);
void sendEnemyPosition(robot_position &position)
{
    cvsysQueue.sendStruct(PM_CVSYS_OPPONENT_POSITION, PRIORITY_HIGH, sizeof(robot_position), (char*) &position);
    aiQueue.sendStruct(PM_AI_OPPONENT_POSITION, PRIORITY_HIGH, sizeof(robot_position), (char*) &position);
}

void sendOurPosition(robot_position &position)
{
    cvsysQueue.sendStruct(PM_CVSYS_ROBOT_POSITION, PRIORITY_HIGH, sizeof(robot_position), (char*) &position);
}

// end communication


static void urg_exit(urg_t *urg, const char *message)
{
    printf("%s: %s\n", message, urg_error(urg));
    urg_disconnect(urg);

    exit(1);
}

int distance2size(int distance, int diameter)
{
    // circumference for circles, radius = distance.
    double circumference = 2 * M_PI * distance;

    // each angle step is 0.36 degrees.
    return diameter*1000/circumference;
}

typedef struct
{
    double angle;
    long distance;
} ang_dist;


int pos_x(double angle, double orientation, int distance, int x_offset)
{
    return sin(angle+orientation)*distance+x_offset;
}

int pos_y(double angle, double orientation, int distance, int y_offset)
{
    return cos(angle+orientation)*distance+y_offset;
}

void corn_draw(RgbImage& input, double scale)
{
    vector< CvPoint > corn;
    CvPoint start;
    corn.assign(18,start);
    start.x=150;
    start.y=740;
    for (int i=0; i<18; i++)
    {
        int n=6;
        if (i<15) n=5;
        if (i<12) n=4;
        if (i<10) n=3;
        if (i<8) n=2;
        if (i<6) n=1;
        if (i<3) n=0;
        corn[i].x=start.x+n*450;
        switch (i)
        {
        case 0:
        case 15:
            corn[i].y=start.y+0*250;
            break;
        case 3:
        case 12:
            corn[i].y=start.y+1*250;
            break;
        case 1:
        case 6:
        case 10:
        case 16:
            corn[i].y=start.y+2*250;
            break;
        case 4:
        case 8:
        case 13:
            corn[i].y=start.y+3*250;
            break;
        case 2:
        case 7:
        case 11:
        case 17:
            corn[i].y=start.y+4*250;
            break;
        case 5:
        case 9:
        case 14:
            corn[i].y=start.y+5*250;
            break;
        default:
            break;
        }
    }
    for (int i=0; i<corn.size(); i++)
    {
        cvCircle(input.raw(),cvPoint(corn[i].x*scale,corn[i].y*scale),25*scale,CV_RGB(255,0,0), -1);
    }
}

typedef struct
{
    int x;
    int y;
    double orientation;
} location;

//! main
int main(int argc, char *argv[])
{
    vector< vector< ang_dist > > objects;
    robot_position pos;
    robot_position our_pos;
    our_pos.y=120;
    our_pos.x=2970;
    our_pos.psi=20*M_PI/180;
    timeval tp;
    location robot;
    robot.x = 2930; // coordinate in mm.
    robot.y = 270; // coordinate in mm.
    robot.orientation = 270*M_PI/180;
    printf("LIDAR - 2D Continous Visualization\n");
    printf("(C) 2010 Kai Hugo Hustoft Endresen.\n");
    cvNamedWindow("LIDAR");
    RgbImage output = cvCreateImage(cvSize(900,630),8,3);

    enum
    {
        CaptureTimes = 10,
    };

    const char device[] = "/dev/ttyACM0"; /* For Linux */

    int data_max;
    long* data;
    int stop=0;
    int timestamp = -1;
    int previous_timestamp;
    int remain_times;
    int scan_msec;
    urg_parameter_t parameter;
    int ret, n, i, key;
    int min_length = 0;
    int max_length = 0;

    /* Connection */
    urg_t urg;
    ret = urg_connect(&urg, device, 115200);
    if (ret < 0)
    {
        urg_exit(&urg, "urg_connect()");
        exit(1);
    }

    /* Reserve for receive buffer */
    data_max = urg_dataMax(&urg);
    data = (long*)malloc(sizeof(long) * data_max);
    if (data == NULL)
    {
        fprintf(stderr, "data_max: %d\n", data_max);
        perror("data buffer");
        exit(1);
    }
    urg_parameters(&urg, &parameter);
    scan_msec = urg_scanMsec(&urg);

    /* Request for MD data */
    /* To get data continuously for more than 100 times, set capture times equal
       to infinity times(UrgInfinityTimes) */
    /* urg_setCaptureTimes(&urg, UrgInfinityTimes); */
    //assert(CaptureTimes < 100);
    urg_setCaptureTimes(&urg, UrgInfinityTimes);

    /* Request for data */
    ret = urg_requestData(&urg, URG_MD, URG_FIRST, URG_LAST);
    if (ret < 0)
    {
        urg_exit(&urg, "urg_requestData()");
    }

    for (i = 0; stop == 0; ++i)
    {
        /* Reception */
        n = urg_receiveData(&urg, data, data_max);
        gettimeofday(&tp,NULL);
        pos.timestamp=tp;
        printf("n = %d\n", n);
        if (n < 0)
        {
            //urg_exit(&urg, "urg_receiveData()");
	    printf("negative response!\n");
            continue;
        }
        else if (n == 0)
        {
            printf("n == 0\n");
            --i;
            continue;
        }

        /* Display the front data with timestamp */
        /* Delay in reception of data at PC causes URG to discard the data which
           cannot be transmitted. This may  results in remain_times to become
           discontinuous */
        previous_timestamp = timestamp;
        timestamp = urg_recentTimestamp(&urg);
        remain_times = urg_remainCaptureTimes(&urg);


        printf("%d/%d: %ld [mm], %d [msec], (%d)\n",
               remain_times, CaptureTimes, data[parameter.area_front_], timestamp,
               timestamp - previous_timestamp);
        printf("%d, %d\n", i, remain_times);

        /* Neglect the distance data if it is less than urg_minDistance() */
        min_length = urg_minDistance(&urg);
        max_length = urg_maxDistance(&urg);
        /*
        for (int j = 0; j < n; ++j)
        {
            int x, y;
            long length = data[j];

            // Neglect the out of range values
            if ((length <= min_length) || (length >= max_length))
            {
                continue;
            }

            x = (int)(length * cos(urg_index2rad(&urg, j)));
            y = (int)(length * sin(urg_index2rad(&urg, j)));

            printf("%d\t%d\t# %d, %ld\n", x, y, j, length);
        }
        */
        wrapper::zero_fill(output.raw());
        int scale=10;
        output[(int) round(robot.y*0.3)][(int) round(robot.x*0.3)].r=255;
        // the LIDAR is limited to 240 degrees, and:
        // tan 30 = opposite/adjacent
        // adjacent/sqrt(3) = opposite.
        int y1 = pos_y(-120*M_PI/180,robot.orientation, 100, robot.y);
        int y2 = pos_y(120*M_PI/180,robot.orientation, 100, robot.y);
        int x1 = pos_x(-120*M_PI/180, robot.orientation, 100, robot.x);
        int x2 = pos_x(120*M_PI/180,robot.orientation, 100, robot.x);
        cvLine(output.raw(),cvPoint(robot.x*0.3,robot.y*0.3),cvPoint(x1*0.3,y1*0.3),cvScalar(255,0,0),1,8);
        cvLine(output.raw(),cvPoint(robot.x*0.3,robot.y*0.3),cvPoint(x2*0.3,y2*0.3),cvScalar(255,0,0),1,8);
        /*
        cvLine(output.raw(),cvPoint(robot.x,robot.y),cvPoint(,robot.y+400/sqrt(3)),cvScalar(255,0,0),1,8);
        cvLine(output.raw(),cvPoint(robot.x,robot.y),cvPoint(800,robot.y+400/sqrt(3)),cvScalar(255,0,0),1,8);
        */
        corn_draw(output,0.3);
        vector<CvPoint> object_readings;
        for (int j =0; j < n; j++)
        {
            long length = data[j];
            int max_error=0;
            if (length > 1000)
            {
                max_error = 0.03*length;
            }
            else
            {
                max_error = 30;
            }
            if (length < 20)
            {
                continue;
            }
            CvPoint p;
            p.y = pos_y(urg_index2rad(&urg, j), robot.orientation, length, robot.y);
            p.x = pos_x(urg_index2rad(&urg, j), robot.orientation, length, robot.x);
            if (p.y>0 && p.x>0 && p.x<3000 && p.y<2100)
            {
                if (p.y<550 && p.x <3000-750)
                {
                    continue;
                }
                else
                {
                    cvCircle(output.raw(),cvPoint(p.x*0.3,p.y*0.3),max_error*0.3, CV_RGB(25,25,25), 1);
                    output[(int) round(p.y*0.3)][(int) round(p.x*0.3)].g=255;
                    object_readings.push_back(p);
                    //cout << "Robot position: " << p.x << " mm, " << p.y << " mm" << endl;
                }
            }
        }
        if(object_readings.size()<1) {
            continue;
        }
        CvPoint p = object_readings.at((object_readings.size()-1)/2);
        cout << "Robot position: " << p.x << " mm, " << p.y << " mm" << endl;
        pos.y=p.y;
        pos.x=p.x;
        pos.psi=90*M_PI/180.0;
	//	if(i % 10 == 0) {
        	sendEnemyPosition(pos);
	//}
        wrapper::show_image("LIDAR",output);
        key = cvWaitKey(20);
        if(key == 'y') {
            our_pos.y-=10;
        }
        if(key == 'h') {
            our_pos.y+=10;
        }
        if(key == 'g') {
            our_pos.x-=10;
        }
        if(key == 'j') {
            our_pos.x+=10;
        }
        if(key == 'u') {
            our_pos.psi-=1*M_PI/180.0;
        }
        if(key == 'i') {
            our_pos.psi+=1*M_PI/180.0;
        }
        printf("POS_NOW: %d %d, angle: %f\n",our_pos.y,our_pos.x,our_pos.psi*180.0/M_PI);
        our_pos.timestamp=pos.timestamp;
        sendOurPosition(our_pos);
        if (key == 'q')
        {
            stop = 1;
        }

        if (remain_times <= 0)
        {
            break;
        }
    }
    printf("Disconnecting..\n");
    urg_disconnect(&urg);
    printf("Freeing..\n");
    free(data);
    printf("All done!\n");
    return 0;
}
